"""
NRV-Cellular Level postprocessing
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import faulthandler
from collections.abc import Iterable

import numpy as np
from numba import jit
from scipy import signal

from ...backend.file_handler import is_iterable, json_dump, json_load
from ...backend.log_interface import pass_info, rise_warning, rise_error
from ..units import *
from ...nmod.unmyelinated import unmyelinated
from ...nmod.myelinated import myelinated
from ...backend.NRV_Class import load_any

# enable faulthandler to ease 'segmentation faults' debug
faulthandler.enable()


#############################
## miscellaneous functions ##
#############################
def distance_point2point(x_1, y_1, x_2=0, y_2=0):
    """
    Computes the distance between a point (x_p,y_p) and a line defined as y=a*x+b

    Parameters
    ----------
    x_1 : float
        first point x coordinate,
    y_1 : float
        firstpoint y coordinate,
    x_2 : float
        second point x coordinate, by default 0
    y_2 : float
        second point y coordinate, , by default 0

    Returns
    --------
    d : float
        distance between the point and the orthogonal projection of (x_p,y_p) on it
    """
    d = ((x_1 - x_2) ** 2 + (y_1 - y_2) ** 2) ** 0.5
    return d


def distance_point2line(x_p, y_p, a, b):
    """
    Computes the distance between a point (x_p,y_p) and a line defined as y=a*x+b

    Parameters
    ----------
    x_p : float
        point x coordinate,
    y_p : float
        point y coordinate,
    a   : float
        line direction coeefficient
    b   : float
        line y for x = 0

    Returns
    --------
    d : float
        distance between the point and the orthogonal projection of (x_p,y_p) on it
    """
    d = np.abs(a * x_p - y_p + b) / (np.sqrt(a**2 + 1))
    return d


#####################################
## SAVE AND LOAD RESULTS FUNCTIONS ##
#####################################


def save_axon_results_as_json(results, filename):
    """
    save dictionary as a json file

    Parameters
    ----------
    results     : dictionary
        stuff to save
    filename    : str
        name of the file where results are saved
    """
    json_dump(results, filename)


def load_simulation_from_json(filename):
    """
    load results of individual axon simulation from a json. This function is specific to axons simulation as some identified results are automatically converted as numpy arrays.

    Parameters
    ----------
    filename    : str
        name of the file where axons simulations are saved
    """
    results = json_load(filename)

    # convert iterables to numpy arrays
    int_iterables = [
        "node_index",
        "Markov_Nav_modeled_NoR",
        "V_mem_raster_position",
        "V_mem_filtered_raster_position",
        "V_mem_raster_time_index",
        "V_mem_filtered_raster_time_index",
    ]
    for key, value in results.items():
        if is_iterable(value) and not isinstance(results[key],dict):
            if key in int_iterables:
                results[key] = np.asarray(value, dtype=np.int16)
            else:
                if len(value) == 0:
                    results[key] = []
                elif isinstance(value[0], str):
                    results[key] = value
                else:
                    if key in "rec_position_list":
                        results[key] = value
                    else:
                        results[key] = np.asarray(value, dtype=np.float32)
        else:
            results[key] = value
    return results


##############################################
## HANDLE THE SIMULATION RESULT DICTIONNARY ##
##############################################
def remove_key(my_dict, key, verbose=False):
    """
    Remove an item from a dictionary, usefull before saving files, as some results maybe heavy and are potentially useless after some steps of postprocessing.

    Parameters
    ----------
    my_dict : dictionary
        dictionary where an item should be deleted
    key     : str
        name of the key to delete
    """
    # if isinstance(key, Iterable):
    #    for k in key:
    #        del my_dict[k]
    # else:
    del my_dict[key]
    pass_info("removed the following key from results: ", key, verbose=verbose)


def remove_non_NoR_zones(my_dict, key):
    """
    Automatically remove values out of nodes of Ranvier for membrane voltage and associated quantities.
    This function is helpfull for large simulation before saving results

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be cleaned
    key     : str
        name of the key to clean
    """
    if "V_mem" in key:
        if my_dict["Axon_type"] == "Myelinated":
            new_entry = []
            for i in my_dict["Nodes_of_Ranvier_indexes"]:
                new_entry.append(my_dict[key][i, :])
            my_dict[key] = np.asarray(new_entry)
        else:
            rise_warning(
                "Warning, remove_non_NoR_zones only applicable to Myelinated axons"
            )
    else:
        rise_warning(
            "Warning, remove_non_NoR_zones only applicable to membrane voltage or current"
        )


def generate_axon_from_results(results):
    """
    generate a blank axon with the same dimensions as in the results

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    ax              : unmyelinated or myelinated
        blank axon with the same dimensions as in the results
    """
    if not results["myelinated"]:
        ax = unmyelinated(**results)
    else:
        ax = myelinated(**results)
    return ax


############################
## AXON SIGNAL PROCESSING ##
############################
def filter_freq(my_dict, my_key, freq, Q=10):
    """
    Basic Filtering of quantities. This function design a notch filter (scipy IIR-notch).
    Adds an item to the specified dictionary, with the key termination '_filtered' concatenated to the original key.

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be filtered
    key     : str
        name of the key to filter
    freq    : float or array, list, np.array
        frequecy or list of frequencies to filter in kHz, as time is defined in ms in NRV2.
        If multiple frequencies, they are filtered sequencially, with as may filters as frequencies, in the specified order
    Q       : float
        quality factor of the filter, by default set to 10
    """
    if isinstance(freq, Iterable):
        f0 = np.asarray(freq)
    else:
        f0 = freq
    if my_dict["dt"] == 0:
        rise_warning(
            "Warning: filtering aborted, variable time step used for differential equation solving"
        )
        return False
    else:
        fs = 1 / my_dict["dt"]
        if isinstance(f0, Iterable):
            new_sig = np.zeros(my_dict[my_key].shape)
            for k in range(len(my_dict[my_key])):
                new_sig[k, :] = my_dict[my_key][k]
                for f in f0:
                    b_notch, a_notch = signal.iirnotch(f, Q, fs)
                    new_sig[k, :] = signal.lfilter(b_notch, a_notch, new_sig[k, :][k])
        else:
            ##  NOTCH at the stimulation frequency
            b_notch, a_notch = signal.iirnotch(f0, Q, fs)
            new_sig = np.zeros(my_dict[my_key].shape)
            for k in range(len(my_dict[my_key])):
                new_sig[k, :] = signal.lfilter(b_notch, a_notch, my_dict[my_key][k])
        my_dict[my_key + "_filtered"] = new_sig


def rasterize(
    my_dict, my_key, t_start=0, t_stop=0, t_min_spike=0.1, t_refractory=2, threshold=0
):
    """
    Rasterize a membrane potential (or filtered or any quantity processed from membrane voltage), with spike detection.
    This function adds 4 items to the dictionnary, with the key termination '_raster_position', '_raster_x_position', '_raster_time_index', '_raster_time' concatenated to the original key.
    These keys correspond to:
    _raster_position    : spike position as the indice of the original key
    _raster_x_position  : spike position as geometrical position in um
    _raster_time_index  : spike time as the indice of the original key
    _raster_time        : spike time as ms

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be rasterized
    key     : str
        name of the key to rasterize
    t_start         : float
        time at which the spike detection should start, in ms. By default 0
    t_stop          : float
        maximum time to apply spike detection, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    t_min_spike     : float
        minimum duration of a spike over its threshold, in ms. By default set to 0.1 ms
    t_refractory    : float
        refractory period for a spike, in ms. By default set to 2 ms.
    threshold       : float
        threshold for spike dection, in mV. If 0 is specified the threshold associated with the axon is automatically chosen. By default set to 0.
        Note that if a 0 value is wanted as threshold, a insignificat value (eg. 1e-12) should be specified.
    """
    if t_stop == 0:
        t_stop = int(my_dict["t_sim"] / my_dict["dt"])
    else:
        t_stop = int(t_stop / my_dict["dt"])
    if threshold == 0:
        thr = my_dict["threshold"]
    else:
        thr = threshold
    ## selecting the list of position considering what has been recorded
    if my_dict["myelinated"] == True:
        if my_dict["rec"] == "all":
            list_to_parse = my_dict["node_index"]
            x = my_dict["x"]
        else:
            list_to_parse = np.arange(len(my_dict["x_rec"]))  # my_dict[my_key]
            x = my_dict["x_rec"]
    else:
        list_to_parse = np.arange(len(my_dict["x_rec"]))  # my_dict[my_key]
        x = my_dict["x_rec"]
    # spike detection
    (
        my_dict[my_key + "_raster_position"],
        my_dict[my_key + "_raster_x_position"],
        my_dict[my_key + "_raster_time_index"],
        my_dict[my_key + "_raster_time"],
    ) = spike_detection(
        my_dict[my_key],
        my_dict["t"],
        x,
        list_to_parse,
        thr,
        my_dict["dt"],
        t_start,
        t_stop,
        t_refractory,
        t_min_spike,
    )


@jit(nopython=True, fastmath=True)
def spike_detection(
    Voltage, t, x, list_to_parse, thr, dt, t_start, t_stop, t_refractory, t_min_spike
):
    """
    Internal use only, spike detection just in time compiled to speed up the process
    """
    raster_position = []
    raster_x_position = []
    raster_time_index = []
    raster_time = []
    # parsing to find spikes
    for i in list_to_parse:
        t_last_spike = t_start - t_refractory
        for j in range(int(t_start * (1 / dt)), t_stop):
            if (
                Voltage[i][j] <= thr
                and Voltage[i][j + 1] >= thr
                and Voltage[i][min((j + int(t_min_spike * (1 / dt))), t_stop)] >= thr
                and (j * dt - t_last_spike) > t_refractory
            ):  # 1st line: threshold crossing, 2nd: minimum time above threshold,3rd: refractory period
                # there was a spike, get time and position
                raster_position.append(i)
                raster_x_position.append(x[i])
                raster_time_index.append(j)
                raster_time.append(t[j])
                # memorize the time in ms, to evaluate refractory period
                t_last_spike = j * dt
    # return results
    return (
        np.asarray(raster_position),
        np.asarray(raster_x_position),
        np.asarray(raster_time_index),
        np.asarray(raster_time),
    )


def find_spike_origin(
    my_dict, my_key=None, t_start=0, t_stop=0, x_min=None, x_max=None
):
    """
    Find the start time and position of a spike or a spike train. Only work on rasterized keys

    Parameters
    ----------
    my_dict : dictionary
        dictionary where the quantity should be rasterized
    key     : str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start : float
        time at which the spikes are processed, in ms. By default 0
    t_stop  : float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    x_min   : float
        minimum position for spike processing, in um. If none is specified, the spike are processed starting at the 0 position. By default set to None.
    x_max   : float
        minimum position for spike processing, in um. If None is specified, spikes are processed on the full axon length . By default set to 0.

    Returns
    -------
    start_time          : float
        first occurance time in ms
    start_x_position    : float
        first occurance position in um
    """
    # define max timing if not already defined
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    # find the best raster plot
    if my_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    else:
        good_key_prefix = my_key
    # get data only in time windows
    sup_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes = np.intersect1d(sup_indexes, inf_indexes)
    good_spike_times = my_dict[good_key_prefix + "_time"][good_indexes]
    good_spike_positions = my_dict[good_key_prefix + "_x_position"][good_indexes]
    # get data only in the z window if applicable
    if x_min != None:
        sup_xmin_indexes = np.where(good_spike_positions > x_min)
    else:
        sup_xmin_indexes = np.arange(len(good_spike_positions))
    if x_max != None:
        inf_xmax_indexes = np.where(good_spike_positions < x_max)
    else:
        inf_xmax_indexes = np.arange(len(good_spike_positions))
    good_x_indexes = np.intersect1d(sup_xmin_indexes, inf_xmax_indexes)
    considered_spike_times = good_spike_times[good_x_indexes]
    considered_spike_positions = good_spike_positions[good_x_indexes]
    # fin the minimum time corresponding to spike initiation
    start_index = np.where(considered_spike_times == np.amin(considered_spike_times))
    start_time = considered_spike_times[start_index]
    start_x_position = considered_spike_positions[start_index]
    return start_time, start_x_position


def find_spike_last_occurance(
    my_dict, my_key=None, t_start=0, t_stop=0, direction="up", x_start=0
):
    """
    Find the last position of a spike occurance for rasterized data

    Parameters
    ----------
    my_dict     : dictionary
        dictionary where the quantity should be rasterized
    key         : str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start     : float
        time at which the spikes are processed, in ms. By default 0
    t_stop      : float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    direction   : str
        Direction of the spike propagation, chose between:
            'up'    -> spike propagating to higher x-coordinate values
            'down'  -> spike propagating to lower x-coordinate values
    x_start     : float
        minimum position for spike processing, in um. If None is specified, spikes are processed on the full axon length . By default set to 0.

    Returns
    -------
    t_last      : float
        last occurance time in ms
    x_last  : float
        last occurance position in um
    """
    # define max timing if not already defined
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    # find the best raster plot
    if my_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    else:
        good_key_prefix = my_key
    # get x_start, eventually t_start
    if x_start == 0:
        t_start, x_start = find_spike_origin(
            my_dict, my_key=good_key_prefix, t_start=t_start, t_stop=t_stop
        )
    # get data only in time windows
    sup_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes = np.intersect1d(sup_indexes, inf_indexes)
    considered_spike_times = my_dict[good_key_prefix + "_time"][good_indexes]
    considered_spike_positions = my_dict[good_key_prefix + "_x_position"][good_indexes]
    if direction == "up":
        # find the spike in the upper region of the start
        upper_spike_index = np.where(considered_spike_positions > x_start)
        upper_spike_times = considered_spike_times[upper_spike_index]
        upper_spike_positions = considered_spike_positions[upper_spike_index]
        # find the last occurance
        last_spike_index = np.where(upper_spike_times == np.amax(upper_spike_times))
        t_last = upper_spike_times[last_spike_index]
        x_last = upper_spike_positions[last_spike_index]
    elif direction == "down":
        # find the spike in the upper region of the start
        lower_spike_index = np.where(considered_spike_positions < x_start)
        lower_spike_times = considered_spike_times[lower_spike_index]
        lower_spike_positions = considered_spike_positions[lower_spike_index]
        # find the last occurance
        last_spike_index = np.where(lower_spike_times == np.amax(lower_spike_times))
        t_last = lower_spike_times[last_spike_index]
        x_last = lower_spike_positions[last_spike_index]
    else:
        # find the spike in the upper region of the start
        upper_spike_index = np.where(considered_spike_positions > x_start)
        upper_spike_times = considered_spike_times[upper_spike_index]
        upper_spike_positions = considered_spike_positions[upper_spike_index]
        # find the spike in the upper region of the start
        lower_spike_index = np.where(considered_spike_positions < x_start)
        lower_spike_times = considered_spike_times[lower_spike_index]
        lower_spike_positions = considered_spike_positions[lower_spike_index]
        # find the last occurances
        last_upper_spike_index = np.where(
            upper_spike_times == np.amax(upper_spike_times)
        )
        last_lower_spike_index = np.where(
            lower_spike_times == np.amax(lower_spike_times)
        )
        t_last = np.asarray(
            [
                lower_spike_times[last_lower_spike_index][0],
                lower_spike_times[last_upper_spike_index],
            ][0]
        )
        x_last = np.asarray(
            [
                lower_spike_positions[last_lower_spike_index][0],
                lower_spike_positions[last_upper_spike_index],
            ][0]
        )
    return t_last, x_last


def speed(my_dict, position_key=None, t_start=0, t_stop=0, x_start=0, x_stop=0):
    """
    Compute the velocity of a spike from rasterized data in a dictionary. The speed can be either positive or negative depending on the propagation direction.

    Parameters
    ----------
    my_dict     : dictionary
        dictionary where the quantity should be rasterized
    key         : str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start     : float
        time at which the spikes are processed, in ms. By default 0
    t_stop      : float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.
    x_start     : float
        minimum position for spike processing, in um. By default set to 0.
    x_stop      : float
        maximum position for spike processing, in um. If 0 is specified, spikes are processed on the full axon length . By default set to 0.

    Returns
    -------
    speed   : float
        velocity.

    Note
    ----
    the velocity is computed with first and last occurance of a spike, be careful specifying the computation window if multiple spikes. This function will not see velocity variation.
    """
    # define max timing if not already defined
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    if t_start == 0:
        if "intra_stim_starts" in my_dict and my_dict["intra_stim_starts"] != []:
            t_start = my_dict["intra_stim_starts"][0]
    if x_start == 0:
        x_stop = my_dict["L"]
    elif x_stop == 0:
        x_start = my_dict["L"]
    # find the best raster plot
    if position_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    else:
        good_key_prefix = position_key
    # get data only in time windows
    sup_time_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_time_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
    sup_position_indexes = np.where(
        my_dict[good_key_prefix + "_x_position"][good_indexes_time] >= x_start
    )
    inf_position_indexes = np.where(
        my_dict[good_key_prefix + "_x_position"][good_indexes_time] <= x_stop
    )
    good_indexes_position = np.intersect1d(sup_position_indexes, inf_position_indexes)
    good_indexes = np.intersect1d(good_indexes_position, good_indexes_time)
    good_spike_times = my_dict[good_key_prefix + "_time"][good_indexes]
    good_spike_positions = my_dict[good_key_prefix + "_x_position"][good_indexes]
    max_time = np.argmax(good_spike_times)
    min_time = np.argmin(good_spike_times)
    speed = (
        (good_spike_positions[max_time] - good_spike_positions[min_time])
        * 10**-3
        / (good_spike_times[max_time] - good_spike_times[min_time])
    )
    return speed


def block(my_dict, position_key=None, t_start=0, t_stop=0):
    """
    check if an axon is blocked or not. The simulation has to include the test spike. This function will look for the test spike initiation and check the propagation

    Parameters
    ----------
    my_dict 	: dictionary
        dictionary where the quantity should be rasterized
    key 		: str
        name of the key to consider, if None is specified, the rasterized is automatically chose with preference for filtered-rasterized keys.
    t_start		: float
        time at which the test spikes can occur, in ms. By default 0
    t_stop		: float
        maximum time at which the spikes are processed, in ms. If zero is specified, the spike detection is applied to the full signal duration. By default set to 0.

    Returns
    -------
    flag 	: bool or None
        True if the axon is blocked, False if not blocked and None if the test spike does not cross the stimulation near point in the simulation (no possibility to check for the axon state)
    """
    position_max = 0
    blocked_spike_positionlist = []
    if t_stop == 0:
        t_stop = my_dict["t_sim"]
    if t_start == 0:
        if "intra_stim_starts" in my_dict and my_dict["intra_stim_starts"] != []:
            t_start = my_dict["intra_stim_starts"][0]

    if position_key == None:
        if "V_mem_filtered_raster_position" in my_dict:
            good_key_prefix = "V_mem_filtered_raster"
        elif "V_mem_raster_position" in my_dict:
            good_key_prefix = "V_mem_raster"
        else:
            # there is no rasterized voltage, nothing to find
            return False
    sup_time_indexes = np.where(my_dict[good_key_prefix + "_time"] > t_start)
    inf_time_indexes = np.where(my_dict[good_key_prefix + "_time"] < t_stop)
    good_indexes_time = np.intersect1d(sup_time_indexes, inf_time_indexes)
    good_spike_times = my_dict[good_key_prefix + "_time"][good_indexes_time]
    blocked_spike_positionlist = my_dict[good_key_prefix + "_x_position"][
        good_indexes_time
    ]
    if len(blocked_spike_positionlist) == 0:
        return None
    if "intra_stim_positions" in my_dict:
        if my_dict["intra_stim_positions"] < my_dict["extracellular_electrode_x"]:
            position_max = max_spike_position(
                blocked_spike_positionlist, position_max, spike_begin="down"
            )

            if blocked_spike_positionlist[position_max] < 9.0 / 10 * my_dict["L"]:
                return True
            else:
                for i in range(position_max - 1):
                    if (
                        blocked_spike_positionlist[i + 1]
                        - blocked_spike_positionlist[i]
                        > my_dict["L"] / 5
                    ):
                        return True
                else:
                    return False
        else:
            position_max = max_spike_position(
                blocked_spike_positionlist, position_max, spike_begin="up"
            )
            if min(blocked_spike_positionlist) > 1.0 / 10 * my_dict["L"]:
                return True
            else:
                for i in range(position_max - 1):
                    if (
                        blocked_spike_positionlist[i]
                        - blocked_spike_positionlist[i + 1]
                        > my_dict["L"] / 5
                    ):
                        return True
                else:
                    return False
    else:
        print("intra_stim_positions is not in dictionnary")


def max_spike_position(blocked_spike_positionlist, position_max, spike_begin="down"):
    if spike_begin == "down":
        while blocked_spike_positionlist[
            position_max + 1
        ] >= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max
    else:
        while blocked_spike_positionlist[
            position_max + 1
        ] <= blocked_spike_positionlist[position_max] and position_max < (
            len(blocked_spike_positionlist) - 2
        ):
            position_max = position_max + 1
        return position_max


@jit(nopython=True, fastmath=True)
def count_spike(onset_position):
    """
    spike counting, just in time compiled. For internal use only.
    """
    if len(onset_position) == 0:
        spike_number = 0
        return 0
    else:
        spike_number = 1
        for i in range(len(onset_position) - 1):
            if onset_position[i] == min(onset_position):
                if onset_position[i] == onset_position[i + 1]:
                    spike_number = spike_number + 1
    return spike_number


def check_test_AP(results_sim):
    """
    Check if a test AP is correctely triggered during an axon simulation and if so return the\
    trigger time

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    test_AP     : float or None
        time in ms of the first test AP during the simulation. None if no test AP found
    """
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)
    if "intra_stim_starts" not in results_sim:
        return None
    else:
        mask = False
        test_AP = results_sim["intra_stim_starts"]
        if len(test_AP):
            if is_iterable(test_AP):
                test_AP = test_AP[0]
            i_first_pos = np.where(results_sim["V_mem_raster_x_position"] == 0)
            for i in i_first_pos[0]:
                if (
                    results_sim["V_mem_raster_time"][i] >= test_AP - 0.01
                    and results_sim["V_mem_raster_time"][i] <= test_AP + 0.7
                ):
                    mask = True
            if not mask:
                test_AP = None
        else:
            test_AP = None
        return test_AP


def detect_start_extrastim(results_sim, threshold=None):
    """
    Returns the starting time of extracellular stimulation from axon simulation results

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method
    threshold       : float
        Current threshold (uA) to consider the stimulation started, if None take the time of second point of the\
        stimulation, by default None

    Returns
    -------
    t_start     : list or float or None
        list of stimulation starting time in ms, float if only one stimulation and None if no stimulation
    """
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    t_start = []
    if "extracellular_stimuli" in results_sim:
        for i in range(len(results_sim["extracellular_stimuli"])):
            s = results_sim["extracellular_stimuli"][i]
            t = results_sim["extracellular_stimuli_t"][i]
            if threshold is None:
                t_start += [t[1]]
            else:
                for i in reversed(range(len(s))):
                    if abs(s[i]) > threshold:
                        t_start += [t[i]]
    if len(t_start) == 0:
        t_start = None
    elif len(t_start) == 1:
        t_start = t_start[0]
    return t_start


def extra_stim_properties(results_sim):
    """
    Return elect caracteristics (blocked, Onset response, ...)

    Parameters
    ----------
    results_sim       : dict or str
        simulation results dictionary or path and name of the file containing it
    Returns
    -------
    electrode       : dict
        dictonry containing the position (x, y, z), the stimulation start (ms) and maximum value (uA)
        for each electrodes
    """
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    electrode = {}

    if "extracellular_electrode_y" in results_sim:
        electrode["x"] = results_sim["extracellular_electrode_x"]
        electrode["y"] = results_sim["extracellular_electrode_y"]
        electrode["z"] = results_sim["extracellular_electrode_z"]
        electrode["start"] = detect_start_extrastim(results_sim)
        electrode["max amplitude"] = [
            max(s) for s in results_sim["extracellular_stimuli"]
        ]

    return electrode


def axon_state(results_sim, save=False, saving_file="axon_state.json"):
    """
    Return axon caracteristics (blocked, Onset response, ...) from axon simulation results

    Parameters
    ----------
    results_sim       : dict or str
        simulation results dictionary or path and name of the saving file
    save        : bool
        if True save result in json file
    saving_file : str
        if save is True path and name of the saving file

    Returns
    -------
    axon_state       : dict
        dictionary containing axon caracteristics
    """
    if type(results_sim) == str:
        results_sim = load_simulation_from_json(results_sim)

    ID = results_sim["ID"]

    # Axon parameter
    parameters = {}

    if "diameter" in results_sim:
        parameters["diameter"] = results_sim["diameter"]

    if "myelinated" in results_sim and results_sim["myelinated"]:
        parameters["node"] = len(results_sim["x_nodes"])

    if (
        "extracellular_electrode_y" in results_sim
        and len(results_sim["extracellular_electrode_y"]) == 1
    ):
        parameters["distance electrod"] = distance_point2line(
            results_sim["y"],
            results_sim["z"],
            results_sim["extracellular_electrode_y"][0],
            results_sim["extracellular_electrode_z"][0],
        )
        if results_sim["myelinated"]:
            x_elec = results_sim["extracellular_electrode_x"][0]
            elec_node = np.argmin(abs(results_sim["x_nodes"] - x_elec))
            elec_ali = (results_sim["x_nodes"][elec_node] - x_elec) / (
                results_sim["x_nodes"][1] - results_sim["x_nodes"][0]
            )
            parameters["electrod node"] = int(elec_node)
            parameters["electrod alignment"] = float(elec_ali)

    # Check Block
    test_AP = check_test_AP(results_sim)
    if test_AP is None:
        block_state = None
    else:
        if "extracellular_electrode_x" not in results_sim:
            results_sim["extracellular_electrode_x"] = 0
        block_state = block(
            results_sim, t_start=test_AP - 0.001
        )  # , t_stop=test_AP+1) # Gerer le delay
    # Check Onset Response

    onset_state = False
    t_start_stim = detect_start_extrastim(results_sim)

    pos = results_sim["V_mem_raster_position"]
    if results_sim["myelinated"]:
        M = len(["x_nodes"])
    else:
        M = len(["x_rec"])
    i_first_pos = np.where(pos == 0)
    i_last_pos = np.where(pos == M)

    # Count Onset response
    N_AP = (len(i_first_pos[0]) + len(i_last_pos[0])) / 2
    if test_AP is not None:
        if block_state:
            N_AP -= 0.5
        else:
            N_AP -= 1

    if N_AP > 0:
        onset_state = True

    axon_state = {
        "ID": ID,
        "parameters": parameters,
        "block_state": block_state,
        "onset_state": onset_state,
        "onset number": N_AP,
    }
    if save:
        save_axon_results_as_json(axon_state, saving_file)

    return axon_state


##############################
## Axon properties function ##
##############################
def get_index_myelinated_sequence(results, n):
    """
    Returns the sequence

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    n_center              : int
        value of the cutoff frequency of the axon's membrane
    """
    if not results["myelinated"] or results["rec"] == "node":
        return "node"
    else:
        if n > len(results["x_rec"]):
            rise_warning("index not in axon")
        # +1 required because nbr of computation point = nbr seg/sec + 1
        # see if it's a bug
        Nseg_per_sec = results["Nseg_per_sec"] + 1
        N_sec_type = 11
        seq_types = results["axon_path_type"]
        if n == 0:
            return seq_types[0]
        else:
            return seq_types[((n - 1) // Nseg_per_sec) % N_sec_type]


def find_central_node_index(results):
    """
    Returns the index of the closer node from the center from a dictionnary results

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    n_center              : int
        value of the cutoff frequency of the axon's membrane
    """
    n_center = len(results["x_rec"]) // 2
    if not results["myelinated"] or results["rec"] == "node":
        return n_center
    else:
        for i in range(n_center):
            if get_index_myelinated_sequence(results, n_center + i) == "node":
                return n_center + i
            elif get_index_myelinated_sequence(results, n_center - i) == "node":
                return n_center - i
    rise_warning("No node found in the axon")
    return n_center


def compute_f_mem(results):
    """
    compute the cutoff frequency of the axon's membrane and add it to the simulation results dictionnary
    NB: The frequency is computed in [kHz]

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    f_mem              : np.ndarray
        value of the cutoff frequency of the axon's membrane
    """

    if "g_mem" not in results:
        rise_warning("f_mem cannot be computed computed without membrane conductivity")
        return None
    if "f_mem" not in results:
        ax = generate_axon_from_results(results)
        results["c_mem"] = ax.get_membrane_capacitance()
        del ax
    N_x, N_t = np.shape(results["g_mem"])
    f_mem = np.zeros((N_x, N_t))
    for i_t in range(N_t):
        f_mem[:, i_t] = results["g_mem"][:, i_t] / (2 * np.pi * results["c_mem"])

    # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
    # * [MHz] to convert to [kHz]
    results["f_mem"] = f_mem * MHz
    return results["f_mem"]

def get_myeline_properties(results):
    """
    compute the cutoff frequency of the axon's membrane and add it to the simulation results dictionnary
    NB: The frequency is computed in [kHz]

    Parameters
    ----------
    results_sim     : dict
        results of axon.simulate method

    Returns
    -------
    f_mem              : np.ndarray
        value of the cutoff frequency of the axon's membrane
    """
    if not results["myelinated"] or results["rec"] == "node":
        rise_warning("No myeline in the axon simulated, None returned")
        return None

    ax = generate_axon_from_results(results)
    results["g_mye"] = ax.get_myeline_conductance()
    results["c_mye"] = ax.get_myeline_capacitance()
    results["f_mye"] = results["g_mye"] / (2 * np.pi * results["c_mye"])

    # in [MHz] as g_mem in [S/cm^{2}] and c_mem [uF/cm^{2}]
    # * [MHz] to convert to [kHz]
    results["f_mem"] *= MHz
    return results["f_mem"]


#############################
## VISUALIZATION FUNCTIONS ##
#############################
def plot_Nav_states(ax, values, title=""):
    """
    Plot the state machine for kinetic (Markov) Nav 1.1 to 1.9 values

    Parameters:
    -----------
    ax      : matplotlib axis object
        axes of the figure to work on
    values  : list, array, numpy array
    """
    states = ["$I_1$", "$I_2$", "$C_1$", "$C_2$", "$O_1$", "$O_2$"]

    X = [-1, -3, 0, 1, 0, 3]
    Y = [0, 0, 1, 0, -1, 0]
    c = ["r", "r", "b", "b", "g", "g"]

    ax.set_xlim(-3.4, 3.4)
    ax.set_ylim(-1.5, 1.5)
    for i in range(len(states)):
        ax.scatter(X[i], Y[i], s=300 + values[i] * 1450, c=c[i], alpha=0.4)
        ax.text(X[i], Y[i], states[i], ha="center", va="center")
    # paths
    ax.arrow(
        -2.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-2, 0.2, "$I_2I_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        -1.5, -0.03, -1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-2, -0.2, "$I_1I_2$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        -0.83, 0.25, 0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(-0.9, 0.6, "$I_1C_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        -0.22,
        0.75,
        -0.5,
        -0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(-0.45, 0.25, "$C_1I_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        0.72, 0.25, -0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.9, 0.6, "$C_1C_2$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        0.33, 0.75, 0.5, -0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.45, 0.25, "$C_2C_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        0.83,
        -0.25,
        -0.5,
        -0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(0.9, -0.6, "$C_2O_1$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        0.22, -0.75, 0.5, 0.5, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(0.45, -0.25, "$O_1C_2$", ha="center", va="center", alpha=0.4)

    ax.arrow(
        -0.33,
        -0.75,
        -0.5,
        0.5,
        linewidth=1,
        alpha=0.5,
        head_width=0.02,
        head_length=0.02,
    )
    ax.text(-0.9, -0.6, "$O_1I_1$", ha="center", va="center", alpha=0.4)

    ax.arrow(1.5, 0.03, 1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02)
    ax.text(2, 0.2, "$C_20_2$", ha="center", va="center", alpha=0.4)
    ax.arrow(
        2.5, -0.03, -1, 0, linewidth=1, alpha=0.5, head_width=0.02, head_length=0.02
    )
    ax.text(2, -0.2, "$O_2C_2$", ha="center", va="center", alpha=0.4)
    # make axes to disappear
    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_title(title)
    ax.axis("off")



"""
NRV-FEM
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import configparser
import os
import time

import numpy as np
from mpi4py import MPI

from ...backend.file_handler import rmv_ext
from ...utils.units import V, mm
from ...backend.MCore import MCH, synchronize_processes
from .fenics_utils.FEMSimulation import FEMSimulation

from .FEM import FEM_model
from .fenics_utils.SimResult import save_sim_res_list
from .mesh_creator.NerveMshCreator import NerveMshCreator, ENT_DOM_offset, pi

# built in FENICS models
dir_path = os.environ["NRVPATH"] + "/_misc"
# material_library = os.listdir(dir_path+"/fenics_templates/")

###############
## Constants ##
###############
machine_config = configparser.ConfigParser()
config_fname = dir_path + "/NRV.ini"
machine_config.read(config_fname)

FENICS_Ncores = int(machine_config.get("FENICS", "FENICS_CPU"))
FENICS_Status = machine_config.get("FENICS", "FENICS_STATUS") == "True"

fem_verbose = True


class FENICS_model(FEM_model):
    """
    A class for FENICS Finite Element Models, inherits from FEM_model.
    """

    def __init__(
        self,
        fname=None,
        Ncore=None,
        handle_server=False,
        elem=None,
        comm="default",
        rank=0,
    ):
        """
        Creates a instance of a FENICS Finite Element Model object.

        Parameters
        ----------
        fname   : str
            path to the mesh file (.msh) file
        Ncore   : int
            number of FENICS cores for computation. If None is specified, this number is taken from the NRV2.ini configuration file. Byu default set to None
        handle_server   : bool
            if True, the instantiation creates the server, else a external server is used. Usefull for multiple cells sharing the same model
        """
        super().__init__(Ncore=Ncore)
        self.type = "FENICS"
        # Default paramerters
        self.L = 5000
        self.y_c = 0
        self.z_c = 0
        self.Outer_D = 5  # mm
        self.Nerve_D = 250  # um
        self.N_fascicle = 0
        self.fascicles = {}
        self.Perineurium_thickness = {}
        self.N_electrode = 0
        self.electrodes = {}
        self.Istim = 1e-3  # A
        self.jstims = []
        self.j_electrode = {}

        self.Outer_mat = "saline"
        self.Epineurium_mat = "epineurium"
        self.Endoneurium_mat = "endoneurium_ranck"
        self.Perineurium_mat = "perineurium"
        self.Electrodes_mat = 1  # "platinum"

        self.default_fascicle = {"D": 200, "y_c": 0, "z_c": 0, "res": 20}
        self.default_electrode = {
            "elec_type": "LIFE",
            "x_c": self.L / 2,
            "y_c": 0,
            "z_c": 0,
            "length": 1000,
            "D": 25,
            "res": 3,
        }

        # Mesh
        self.mesh_file = fname
        self.mesh = None
        self.elem = elem
        if elem is None:
            self.elem = ("Lagrange", 2)

        # FEM
        self.sim = None
        self.sim_res = []

        # Mcore attributes
        if self.Ncore is None:
            self.Ncore = FENICS_Ncores
        self.comm = None
        self.set_Ncore(self.Ncore)
        if comm != "default":
            self.comm = comm
        self.rank = rank

        # Status
        self.mesh_file_status = not self.mesh_file is None
        self.is_sim_ready = False
        self.is_meshed = False
        self.is_computed = False

        if not self.mesh_file_status:
            self.mesh = NerveMshCreator(
                Length=self.L,
                Outer_D=self.Outer_D,
                Nerve_D=self.Nerve_D,
                y_c=self.y_c,
                z_c=self.z_c,
            )

    def save(self, save=False, fname="Fenics_model.json", blacklist=[], **kwargs):
        """
        Save the changes to the model file. (Avoid for the overal weight of the package)
        """
        bl = [i for i in blacklist]
        bl += ["comm", "rank", "mesh", "sim", "sim_res"]
        return super().save(save=save, fname=fname, blacklist=bl, **kwargs)

    def save_results(self, save=False, fname="Fenics_model.json"):
        """
        Save the changes to the model file. (Avoid for the overal weight of the package)
        """
        if self.is_computed:
            save_sim_res_list(self.sim_res, fname)

        elif self.is_meshed:
            self.mesh_file = rmv_ext(fname)
            self.mesh.save(fname)

    def load(self, data="extracel_context.json", **kwargs):
        """
        Load all extracellular context properties from a dictionary or a json file

        Parameters
        ----------
        data    : str or dict
            json file path or dictionary containing extracel_context information
        """
        bl = ["comm", "rank", "mesh", "sim", "sim_res"]
        super().load(data=data, blacklist=bl, **kwargs)
        param = self.get_parameters()
        self.load_from_parameters(**param)

    def load_from_parameters(self, **param):
        """ """
        self.reset_parameters()
        self.reshape_outerBox(param["Outer_D"])
        self.reshape_nerve(
            Nerve_D=param["Nerve_D"],
            Length=param["L"],
            y_c=param["y_c"],
            z_c=param["z_c"],
        )
        for id in param["fascicles"]:
            fasc = param["fascicles"][id]
            per_th = param["Perineurium_thickness"][id]
            self.reshape_fascicle(
                fasc["D"],
                y_c=fasc["y_c"],
                z_c=fasc["z_c"],
                ID=int(id),
                Perineurium_thickness=per_th,
                res=fasc["res"],
            )
        for id in param["electrodes"]:
            elec = param["electrodes"][id]
            self.add_electrode(
                elec_type=elec["type"], ID=int(id), res=elec["res"], **elec["kwargs"]
            )

    #############################
    ## Access model parameters ##
    #############################
    def set_Ncore(self, N):
        """
        Set the number of cores to use for the FEM

        Parameters
        ----------
        N       : int
            Number of cores to set
        """
        super().set_Ncore(N)
        if self.is_multi_proc:
            self.comm = MPI.COMM_WORLD
        else:
            self.comm = MPI.COMM_SELF

    def get_parameters(self):
        """
        Get the  all the parameters in the model as a python dictionary.

        Returns
        -------
        param    : dict
            all parameters as dictionnaries in a list, names a keys, with corresponding values
        """
        param = {}
        param["L"] = self.L
        param["y_c"] = self.y_c
        param["z_c"] = self.z_c
        param["Outer_D"] = self.Outer_D / mm
        param["Nerve_D"] = self.Nerve_D
        param["N_fascicle"] = self.N_fascicle
        param["fascicles"] = self.fascicles
        param["Perineurium_thickness"] = self.Perineurium_thickness
        param["N_electrode"] = self.N_electrode
        param["electrodes"] = self.electrodes
        param["Outer_mat"] = self.Outer_mat
        param["Epineurium_mat"] = self.Epineurium_mat
        param["Endoneurium_mat"] = self.Endoneurium_mat
        param["Perineurium_mat"] = self.Perineurium_mat
        param["Perineurium_mat"] = self.Perineurium_mat
        param["Istim"] = self.Istim
        return param

    def __update_parameters(self, bcast=False):
        """
        Internal use only: updates all the parameters from the mesh

        Parameters
        ----------
        bcast       : bool
            if true the parameters are updated on all process
        """
        if MCH.do_master_only_work():
            if bcast:
                self.__update_parameters(bcast=False)
                param = self.get_parameters()
            else:
                param = self.mesh.get_parameters()
        else:
            param = None
        if bcast:
            param = MCH.master_broadcasts_array_to_all(param)
        if MCH.do_master_only_work() ^ bcast:
            for key in param:
                self.__dict__[key] = param[key]

    def reset_parameters(self):
        """
        reset model parameters
        """
        self.L = 5000
        self.y_c = 0
        self.z_c = 0
        self.Outer_D = 5  # mm
        self.Nerve_D = 250  # um
        self.N_fascicle = 0
        self.fascicles = {}
        self.Perineurium_thickness = {}
        self.N_electrode = 0
        self.electrodes = {}
        self.Istim = 1e-3  # A
        self.jstims = []
        self.j_electrode = {}

        self.mesh_file_status = self.mesh_file is not None
        self.is_sim_ready = False
        self.is_meshed = False
        self.is_computed = False
        del self.mesh
        if not self.mesh_file_status:
            if MCH.do_master_only_work:
                self.mesh = NerveMshCreator(
                    Length=self.L,
                    Outer_D=self.Outer_D,
                    Nerve_D=self.Nerve_D,
                    y_c=self.y_c,
                    z_c=self.z_c,
                )
            else:
                self.mesh = None

    #####################
    ## customize model ##
    #####################
    def reshape_outerBox(self, Outer_D, res="default"):
        """
        Reshape the size of the FEM simulation outer box

        Parameters
        ----------
        outer_D : float
            FEM simulation outer box diameter, in mm, WARNING, this is the only parameter in mm !
        """
        if not self.mesh_file_status:
            self.mesh.reshape_outerBox(Outer_D=Outer_D, res=res)
            self.__update_parameters()

    def reshape_nerve(self, Nerve_D, Length, y_c=0, z_c=0, res="default"):
        """
        Reshape the nerve of the FEM simulation

        Parameters
        ----------
        Nerve_D                 : float
            Nerve diameter, in um
        Length                  : float
            Nerve length, in um
        y_c                     : float
            Nerve center y-coordinate in um, 0 by default
        z_c                     : float
            Nerve z-coordinate center in um, 0 by default
        Perineurium_thickness   :float
            Thickness of the Perineurium sheet surounding the fascicles in um, 5 by default
        """
        if not self.mesh_file_status:
            self.L = Length
            self.Nerve_D
            self.mesh.reshape_nerve(
                Nerve_D=Nerve_D, Length=Length, y_c=y_c, z_c=z_c, res=res
            )
            self.__update_parameters()

    def reshape_fascicle(
        self, Fascicle_D, y_c=0, z_c=0, ID=None, Perineurium_thickness=5, res="default"
    ):
        """
        Reshape a fascicle of the FEM simulation

        Parameters
        ----------
        Fascicle_D  : float
            Fascicle diameter, in um
        y_c         : float
            Fascicle center y-coodinate in um, 0 by default
        z_c         : float
            Fascicle center y-coodinate in um, 0 by default
        ID          : int
            If the simulation contains more than one fascicles, ID number of the fascicle to reshape as in FENICS
        """
        if not self.mesh_file_status:
            self.mesh.reshape_fascicle(Fascicle_D, y_c, z_c, ID, res)
            self.__update_parameters()
            if ID is None:
                if self.Perineurium_thickness == {}:
                    ID = 0
                else:  ## To check when not all ID are fixed
                    ID = max(self.Perineurium_thickness) + 1
            self.Perineurium_thickness[ID] = Perineurium_thickness

    def remove_fascicles(self, ID=None):
        """
        remove a fascicle of the FEM simulation

        Parameters
        ----------
        ID          : int, None
            ID number of the fascicle to remove, if None, remove all fascicles, by default None
        """
        if ID is None:
            self.fascicles = {}
            self.Perineurium_thickness = {}
            self.N_fascicle = 0
        elif ID in self.fascicles:
            del self.fascicles[ID]
            del self.Perineurium_thickness[ID]
            self.N_fascicle -= 1
        self.mesh.remove_fascicles(ID=ID)

    def add_electrode(self, elec_type, ID=None, res="default", **kwargs):
        """
        TO BE WRITTEN
        """
        if not self.mesh_file_status:
            self.mesh.add_electrode(elec_type=elec_type, ID=ID, res=res, **kwargs)

    ######################
    ## setup simulation ##
    ######################
    def setup_simulations(self, comm=None, rank=None):
        """
        TO BE WRITTEN
        """
        if comm is not None:
            self.comm = comm
        if rank is not None:
            self.rank = rank
        if not self.is_meshed:
            self.build_and_mesh()
        if not self.is_sim_ready and (MCH.do_master_only_work() or self.is_multi_proc):
            t0 = time.time()
            # SETTING DOMAINS
            # del self.mesh
            self.sim = FEMSimulation(
                mesh_file=self.mesh_file,
                mesh=self.mesh,
                elem=self.elem,
                comm=self.comm,
                rank=self.rank,
            )
            # Outerbox domain
            self.sim.add_domain(
                mesh_domain=ENT_DOM_offset["Outerbox"], mat_pty=self.Outer_mat
            )
            # Nerve domain
            self.sim.add_domain(
                mesh_domain=ENT_DOM_offset["Nerve"], mat_pty=self.Epineurium_mat
            )
            for i in self.fascicles.keys():
                self.sim.add_domain(
                    mesh_domain=ENT_DOM_offset["Fascicle"] + (2 * i),
                    mat_pty=self.Endoneurium_mat,
                )
            for _, (i, elec) in enumerate(self.electrodes.items()):
                if not elec["type"] == "LIFE":
                    self.sim.add_domain(
                        mesh_domain=ENT_DOM_offset["Electrode"] + (2 * i),
                        mat_pty=self.Electrodes_mat,
                    )
            # SETTING INTERNAL BOUNDARY CONDITION (for perineuriums)
            for i in self.fascicles:
                thickness = self.Perineurium_thickness[i]
                f_dom = ENT_DOM_offset["Fascicle"] + (2 * i)
                self.sim.add_inboundary(
                    mesh_domain=ENT_DOM_offset["Surface"] + f_dom,
                    mat_pty=self.Perineurium_mat,
                    thickness=thickness,
                    in_domains=[f_dom],
                )
            # SETTING BOUNDARY CONDITION
            # Ground (to the external ring of Outerbox)
            self.sim.add_boundary(mesh_domain=1, btype="Dirichlet", value=0)
            # Injected current from active electrode
            # For EIT change in for E in elec_patren:
            for _, (E, active_elec) in enumerate(self.electrodes.items()):
                E_var = "E" + str(E)
                mesh_domain_3D = self.__find_elec_subdomain(active_elec)
                self.jstims += {self.__find_elec_jstim(active_elec)}
                self.j_electrode[E_var] = 0
                e_dom = (
                    ENT_DOM_offset["Surface"] + ENT_DOM_offset["Electrode"] + (2 * E)
                )
                self.sim.add_boundary(
                    mesh_domain=e_dom,
                    btype="Neuman",
                    variable=E_var,
                    mesh_domain_3D=mesh_domain_3D,
                )
            # set a parallelizable preconditionner if sim solve on multiple precesses
            if self.is_multi_proc:
                self.sim.set_solver_opt(pc_type="hypre")
            self.is_sim_ready = True
            self.preparing_timer += time.time() - t0

    def set_materials(
        self,
        Outer_mat=None,
        Epineurium_mat=None,
        Endoneurium_mat=None,
        Perineurium_mat=None,
        Electrodes_mat=None,
    ):
        """
        Set material files for any domain
        Outer_mat       :str
            Outer box material fname if None not changed, by default None
        Epineurium_mat      :str
            Epineurium material fname if None not changed, by default None
        Endoneurium_mat     :str
            Endoneurium material fname if None not changed, by default None
        Perineurium_mat     :str
            Outer material fname if None not changed, by default None
        Electrodes_mat      :str
            Electrodes material fname if None not changed, by default None
        """
        if Outer_mat is not None:
            self.Outer_mat = Outer_mat
        if Epineurium_mat is not None:
            self.Epineurium_mat = Epineurium_mat
        if Endoneurium_mat is not None:
            self.Endoneurium_mat = Endoneurium_mat
        if Perineurium_mat is not None:
            self.Perineurium_mat = Perineurium_mat
        if Electrodes_mat is not None:
            self.Electrodes_mat = Electrodes_mat

    def __find_elec_subdomain(self, elec) -> int:
        """
        Internal use only:
        """
        if elec["type"] == "LIFE":
            y_e, z_e = elec["kwargs"]["y_c"], elec["kwargs"]["z_c"]
            for i in self.fascicles:
                fascicle = self.fascicles[i]
                d_f, y_f, z_f = fascicle["D"], fascicle["y_c"], fascicle["z_c"]
                if (y_e - y_f) ** 2 + (z_e - z_f) ** 2 < (d_f / 2) ** 2:
                    return 10 + 2 * i
        return 0

    def __find_elec_jstim(self, elec, I=None) -> float:
        """
        Internal use only: return electrical current density in electrode
        from current valeu and electrode geometry
        """
        # Unitary stimulation
        if I is not None:
            self.Istim = I

        if elec["type"] == "CUFF":
            d_e = self.Nerve_D + 2 * elec["kwargs"]["contact_thickness"]
            l_e = elec["kwargs"]["contact_length"]

        elif elec["type"] == "LIFE":
            d_e = elec["kwargs"]["D"]
            l_e = elec["kwargs"]["length"]

        S = pi * (d_e) * (l_e)
        jstim = self.Istim / S
        return jstim

    ###################
    ## Use the model ##
    ###################

    def get_meshes(self):
        """
        Visualize the model mesh
        WARNING: debbug use only + might not work on all os
        """
        if self.is_meshed:
            self.mesh.visualize()

    def build_and_mesh(self):
        """
        Build the geometry and perform meshing process
        """
        if not self.mesh_file_status and not self.is_meshed:
            t0 = time.time()
            self.__update_parameters()
            if self.N_fascicle == 0:
                self.reshape_fascicle(
                    Fascicle_D=self.default_fascicle["D"],
                    y_c=self.default_fascicle["y_c"],
                    z_c=self.default_fascicle["z_c"],
                    res=self.default_fascicle["res"],
                )
            if self.N_electrode == 0:
                self.add_electrode(
                    elec_type=self.default_electrode["elec_type"],
                    x_c=self.default_electrode["x_c"],
                    y_c=self.default_electrode["y_c"],
                    z_c=self.default_electrode["z_c"],
                    length=self.default_electrode["length"],
                    D=self.default_electrode["D"],
                    res=self.default_electrode["res"],
                )
            self.__update_parameters(bcast=self.is_multi_proc)
            if MCH.do_master_only_work():
                self.mesh.compute_mesh()
            if self.is_multi_proc:
                synchronize_processes()
            self.is_meshed = True
            self.mesh.get_info(verbose=True)
            self.meshing_timer += time.time() - t0

    def solve(self, comm=None, rank=None):
        """
        Solve the model
        """
        if not self.is_sim_ready:
            self.setup_simulations()
        if not self.is_computed:
            t0 = time.time()
            # For EIT change in for E in elec_patren:
            for E in range(self.N_electrode):
                for i_elec in self.j_electrode:
                    if i_elec == "E" + str(E):
                        self.j_electrode[i_elec] = self.jstims[E]
                    else:
                        self.j_electrode[i_elec] = 0
                self.sim.prepare_sim(**self.j_electrode)
                self.sim_res += [self.sim.solve()]
                self.sim_res[-1].vout.label = "E" + str(E)
            self.is_computed = True
            self.solving_timer += time.time() - t0

    ######################
    ## results handling ##
    ######################

    def get_potentials(self, x, y, z, E=-1):
        """
        Get the potential on a line to get extracellular potential for axons stimulation.

        Parameters
        ----------
        x   : np.array
            array of x coordinates in the model
        y   : float
            y-coordinate of the axon
        z   : float
            z-coordinate of the axon
        E   : int
            ID of the electrode from witch the potentials should be evaluated

        Returns:
        --------
        array
            All potential for all paramtric sweeps (all electrodes in NRV2 models)\
            (line: electrode selection, column: potential)
        """
        if self.is_computed:
            t0 = time.time()
            line = [[x_, y, z] for x_ in x]
            if self.N_electrode == 1 or (E < self.N_electrode and E >= 0):
                potentials = self.sim_res[E].eval(line, self.is_multi_proc) * V
            else:
                potentials = []
                for E in range(self.N_electrode):
                    potentials += [self.sim_res[E].eval(line, self.is_multi_proc)]
                potentials = np.transpose(np.array(potentials) * V)
            self.access_res_timer += time.time() - t0
            return potentials



"""
NRV-unmyelinated
Authors: Florian Kolbl / Roland Giraud / Louis Regnacq / Thomas Couppey
(c) ETIS - University Cergy-Pontoise - CNRS
"""
import math

import numpy as np

from .axons import *


class unmyelinated(axon):
    """
    Unmyelineated axon class. Automatic refinition of all neuron sections and properties. User-friendly object including model definition
    Inherit from axon class. see axon for further detail.
    """

    def __init__(
        self,
        y=0,
        z=0,
        d=1,
        L=1000,
        model="Rattay_Aberham",
        dt=0.001,
        Nrec=0,
        Nsec=1,
        Nseg_per_sec=0,
        freq=100,
        freq_min=0,
        mesh_shape="plateau_sigmoid",
        alpha_max=0.3,
        d_lambda=0.1,
        v_init=None,
        T=None,
        ID=0,
        threshold=-40,
        **kwarks,
    ):
        """
        initialisation of an unmyelinted axon

        Parameters
        ----------
        y               : float
            y coordinate for the axon, in um
        z               : float
            z coordinate for the axon, in um
        d               : float
            axon diameter, in um
        L               : float
            axon length along the x axins, in um
        model           : str
            choice of conductance based model, possibly:
                "HH"                : original squid giant axon model, warning - low temperature model, not adapted to mamalian modeling
                "Rattay_Aberham"    : Rattay Aberham model, see [1] for details
                "Sundt"             : Sundt model, see [1] for details
                "Tigerholm"         : Tigerholm model, see [1] for details
                "Schild_94"         : Schild 1994 model, see [1] for details
                "Schild_97"         : Schild 1997 model, see [1] for details
        dt              : float
            computation step for simulations, in ms. By default equal to 1 us
        Nrec            : int
            Number of points along the axon to record for simulation results. Between 0 and the number of segment, if set to 0, all segments are recorded
        Nsec            : int
            Number of sections in the axon, by default 1. Usefull to create umnyelinated axons with a variable segment density
        Nseg_per_sec    : int
            Number of segment per section in the axon. If set to 0, the number of segment is automatically computed using d-lambda rule and following paramters. If set by user, please use odd numbers
        freq            : float
            Frequency used for the d-lmbda rule, corresponding to the maximum membrane current frequency, by default set to 100 Hz
        freq_min        : float
            Minimal frequency fot the d-lambda rule when using an irregular number of segment along the axon, if set to 0, all sections have the same frequency determined by the previous parameter
        mesh_shape      : str
            Shape of the frequencial distribution for the dlmabda rule along the axon, pick between:
                "pyramidal"         -> min frequencies on both sides and linear increase up to the middle at the maximum frequency
                "sigmoid"           -> same a befor with sigmoid increase instead of linear
                "plateau"           -> sale as pyramidal except the max frequency is holded on a central plateau
                "plateau_sigmoid"   -> same as previous with sigmoid increase
        alpha_max       : float
            Proportion of the axon set to the maximum frequency for plateau shapes, by default set to 0.3
        d_lambda        : float
            value of d-lambda for the dlambda rule,
        v_init          : float
            Initial value of the membrane voltage in mV, set None to get an automatically model attributed value
        T               : float
            temperature in C, set None to get an automatically model attributed value
        ID              : int
            axon ID, by default set to 0,
        threshold       : float
            voltage threshold in mV for further spike detection in post-processing, by defautl set to -40mV, see post-processing files for further help

        Note
        ----
        reference [1] corresponds to:
            Pelot, N. A., Catherall, D. C., Thio, B. J., Titus, N. D., Liang, E. D., Henriquez, C. S., & Grill, W. M. (2021). Excitation properties of computational models of unmyelinated peripheral axons. Journal of neurophysiology, 125(1), 86-104.
        """
        super().__init__(
            y,
            z,
            d,
            L,
            dt=dt,
            Nseg_per_sec=Nseg_per_sec,
            freq=freq,
            freq_min=freq_min,
            mesh_shape=mesh_shape,
            alpha_max=alpha_max,
            d_lambda=d_lambda,
            v_init=v_init,
            T=T,
            ID=ID,
            threshold=threshold,
            **kwarks,
        )
        self.Nsec = Nsec
        self.Nrec = Nrec
        self.myelinated = False
        if model in unmyelinated_models:
            self.model = model
        else:
            self.model = "Rattay_Aberham"
        self.__compute_axon_parameters()

    def __compute_axon_parameters(self):
        """
        generate axon from parameters set by user
        """
        ## Handling v_init
        if self.v_init is None:
            # model driven
            if self.model == "HH":
                self.v_init = -67.5
            elif self.model == "Rattay_Aberham":
                self.v_init = -70
            elif self.model == "Sundt":
                self.v_init = -60
            elif self.model == "Tigerholm":
                self.v_init = -62  # -55 in Pelot 2020, changed by FK 19/01/2021
            else:
                self.v_init = -70
        else:
            # user driven
            self.v_init = self.v_init
        ## Handling temperature
        if self.T is None:
            # model driven
            if self.model == "HH":
                self.T = 32  # original HH model, cold model, maximal temperature at which spike propagation is not altered
            else:
                self.T = 37  # mamalian models

        # create and connect (if more than 1) sections
        self.unmyelinated_sections = [
            neuron.h.Section(name="U_axon[%d]" % i) for i in range(self.Nsec)
        ]
        for sec in self.unmyelinated_sections:
            # morphologic parameters
            sec.L = self.L / self.Nsec
            sec.diam = self.d
        if self.Nsec > 1:
            for i in range(self.Nsec - 1):
                self.unmyelinated_sections[i + 1].connect(
                    self.unmyelinated_sections[i], 1, 0
                )
        # implement neuron mechanisms
        self.__set_model(self.model)
        # define the geometry of the axon
        self._axon__define_shape()
        # define the number of segments
        self.__set_Nseg()
        # get nodes positions
        self.__get_seg_positions()
        self.__get_rec_positions(self.Nrec)

    def save(
        self,
        save=False,
        fname="axon.json",
        extracel_context=False,
        intracel_context=False,
        rec_context=False,
        blacklist=[],
    ):
        """
        Return axon as dictionary and eventually save it as json file

        Parameters
        ----------
        save    : bool
            if True, save in json files
        fname   : str
            Path and Name of the saving file, by default "axon.json"

        Returns
        -------
        ax_dic : dict
            dictionary containing all information
        """
        blacklist += ["unmyelinated_sections"]
        return super().save(
            save=save,
            fname=fname,
            extracel_context=extracel_context,
            intracel_context=intracel_context,
            rec_context=rec_context,
            blacklist=blacklist,
        )

    def __set_Nseg(self):
        """
        sets the number of segment, either with the number chosen by user,
        or using the dlambda rule, or using the dlambda rule with a frequency array
        of different shapes (see create_Nseg_freq_shape() help for more details on shapes)
        """
        if self.freq_min == 0 or self.Nsec == 1:  # uniform meshing
            # compute the correct number of segments if needed, or take the one given by user
            if self.Nseg_per_sec != 0:
                Nseg = self.Nseg_per_sec
            else:
                Nseg = d_lambda_rule(
                    self.unmyelinated_sections[0].L,
                    self.d_lambda,
                    self.freq,
                    self.unmyelinated_sections[0],
                )
            # set the number of segment for all declared sections
            for sec in self.unmyelinated_sections:
                sec.nseg = Nseg
                if self.model == "Schild_94" or self.model == "Schild_97":
                    sec.nseg_caintscale = Nseg
                    sec.nseg_caextscale = Nseg
                self.Nseg += Nseg
        else:  # non uniform meshing
            freqs = create_Nseg_freq_shape(
                self.Nsec, self.mesh_shape, self.freq, self.freq_min, self.alpha_max
            )
            for k in range(self.Nsec):
                Nseg = d_lambda_rule(
                    self.unmyelinated_sections[k].L,
                    self.d_lambda,
                    freqs[k],
                    self.unmyelinated_sections[k],
                )
                self.unmyelinated_sections[k].nseg = Nseg
                if self.model == "Schild_94" or self.model == "Schild_97":
                    sec.nseg_caintscale = Nseg
                    sec.nseg_caextscale = Nseg
                self.Nseg += Nseg

    def __get_seg_positions(self):
        """
        get all the computation position of the axon. This corresponds to the positions of the segment
        without the duplicates of connected sections
        """
        x_offset = 0
        x = []
        for sec in self.unmyelinated_sections:
            for seg in sec.allseg():
                if x == []:
                    x.append(seg.x * (sec.L) + x_offset)
                else:
                    x_seg = seg.x * (sec.L) + x_offset
                    if x_seg != x[-1]:
                        x.append(x_seg)
            x_offset += sec.L
        self.x = np.asarray(x)

    def __get_rec_positions(self, Nrec):
        """
        get recording x-coordinates and relative positions, for internal use only

        Parameters
        ----------
        Nrec    : int
            number of points to record in the axon during the simulation, can be chosen between 3 and maximum the number of programmed segments + 2 (both sides)
            or will be by default set to the nearest value
        """
        self.rec_position_list = []
        for k in range(self.Nsec):
            self.rec_position_list.append([])
        if Nrec < 4 and Nrec != 0:
            self.Nrec = 3  # at least, 3 positions will be memorized along the axon (extrema and middle)
            self.rec_position_list[0].append(0)
            if (self.Nsec % 2) == 0:
                self.rec_position_list[self.Nsec / 2].append(0)
            else:
                self.rec_position_list[math.floor(self.Nsec / 2)].append(0.5)
            self.rec_position_list[-1].append(1)
            self.x_rec = np.array([0, self.L / 2, self.L])
        elif Nrec == 0 or Nrec > self.Nsec + self.Nseg:  # record on all nodes
            self.Nrec = self.Nsec + self.Nseg + 1
            for k in range(self.Nsec):
                for seg in self.unmyelinated_sections[k].allseg():
                    self.rec_position_list[k].append(seg.x)
                # delete last position as it will be a dupplicate with next section first seg
                del self.rec_position_list[k][-1]
            # for the last section only, add the max position
            self.rec_position_list[-1].append(1)
            self.x_rec = self.x
        else:
            self.Nrec = Nrec
            self.rec_position_list[0].append(0)
            remaining_recs = np.arange(1, self.Nrec - 1) * (self.Nsec / (self.Nrec - 1))
            for rec in remaining_recs:
                self.rec_position_list[math.floor(rec)].append(rec - math.floor(rec))
            self.rec_position_list[-1].append(1)
            self.x_rec = np.linspace(0, self.L, num=self.Nrec, endpoint=True)

    def __set_model(self, model):
        """
        Adds the passive, Hodgking Huxley and extracellular mechanisms,
        set values to to one given by user at the initialisation or default ones. For internal use only.
        """
        for sec in self.unmyelinated_sections:
            # insert mechanisms
            if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
                sec.insert("pas")
            sec.insert("extracellular")
            sec.xg[0] = 1e10  # short circuit, no myelin
            sec.xc[0] = 0  # short circuit, no myelin
            ## Except for HH, the following code is directly take from modelDB: https://senselab.med.yale.edu/ModelDB/showmodel.cshtml?model=266498#tabs-1
            ## Pelot N (2020) Excitation Properties of Computational Models of Unmyelinated Peripheral Axons J Neurophysiology
            if model == "HH":
                sec.insert("hh")
                # explicit mechanisms settings
                sec.cm = 1.0
                sec.Ra = 200.0
                sec.gnabar_hh = 0.120
                sec.gkbar_hh = 0.036
                sec.gl_hh = 0.0003
                sec.ena = 50
                sec.ek = -77
                sec.el_hh = -54.3
            elif model == "Rattay_Aberham":
                sec.insert(
                    "RattayAberham"
                )  # Model adjusted for a resting potential of -70mV instead of 0 (subtract Vrest from each reversal potential)
                sec.Ra = 100
                sec.cm = 1
                sec.v = -70
                sec.ena = 45
                sec.ek = -82
                sec.e_pas = -70
            elif model == "Sundt":
                sec.insert("nahh")
                sec.gnabar_nahh = 0.04
                sec.mshift_nahh = -6  # NaV1.7/1.8 channelshift
                sec.hshift_nahh = 6  # NaV1.7/1.8 channelshift

                sec.insert("borgkdr")  # insert delayed rectifier K channels
                sec.gkdrbar_borgkdr = 0.04  # density of K channels
                sec.ek = -90  # K equilibrium potential

                sec.g_pas = 1 / 10000  # set Rm = 10000 ohms-cm2
                sec.Ra = 100  # intracellular resistance
                sec.v = -60
                sec.e_pas = (
                    sec.v + (sec.ina + sec.ik) / sec.g_pas
                )  # calculate leak equilibrium potential
            elif self.model == "Tigerholm":
                sec.insert("ks")
                sec.gbar_ks = 0.0069733
                sec.insert("kf")
                sec.gbar_kf = 0.012756
                sec.insert("h")
                sec.gbar_h = 0.0025377
                sec.insert("nattxs")
                sec.gbar_nattxs = 0.10664
                sec.insert("nav1p8")
                sec.gbar_nav1p8 = 0.24271
                sec.insert("nav1p9")
                sec.gbar_nav1p9 = 9.4779e-05
                sec.insert("nakpump")
                sec.smalla_nakpump = -0.0047891
                sec.insert("kdrTiger")
                sec.gbar_kdrTiger = 0.018002
                sec.insert("kna")
                sec.gbar_kna = 0.00042
                sec.insert("naoi")

                sec.insert("koi")
                sec.theta_naoi = 0.029
                sec.theta_koi = 0.029

                sec.insert("leak")
                sec.insert("extrapump")

                sec.Ra = 35.5
                sec.cm = 1.0

                sec.celsiusT_ks = self.T
                sec.celsiusT_kf = self.T
                sec.celsiusT_h = self.T
                sec.celsiusT_nattxs = self.T
                sec.celsiusT_nav1p8 = self.T
                sec.celsiusT_nav1p9 = self.T
                sec.celsiusT_nakpump = self.T
                sec.celsiusT_kdrTiger = self.T
                sec.v = -55
            else:  # Schild 94 or 97 models
                R = 8314
                F = 96500
                sec.insert(
                    "leakSchild"
                )  # All mechanisms from Schild 1994 inserted into model
                sec.insert("kd")
                sec.insert("ka")
                sec.insert("can")
                sec.insert("cat")
                sec.insert("kds")
                sec.insert("kca")
                sec.insert("caextscale")
                sec.insert("caintscale")
                sec.insert("CaPump")
                sec.insert("NaCaPump")
                sec.insert("NaKpumpSchild")
                if self.model == "Schild_94":
                    sec.insert("naf")
                    sec.insert("nas")
                else:
                    sec.insert("naf97mean")
                    sec.insert("nas97mean")
                # Ionic concentrations
                # cao0_ca_ion = 2.0                                      # not in section, adapter considering: https://neuronsimulator.github.io/nrn/rxd-tutorials/initialization.html
                neuron.h.cao0_ca_ion = 2.0  # [mM] Initial Cao Concentration
                # cai0_ca_ion = 0.000117                                 # same as cao_ca_ion
                neuron.h.cai0_ca_ion = 0.000117  # [mM] Initial Cai Concentrations
                ko = 5.4  # [mM] External K Concentration
                ki = 145.0  # [mM] Internal K Concentration
                kstyle = neuron.h.ion_style(
                    "k_ion", 1, 2, 0, 0, 0
                )  # Allows ek to be calculated manually
                sec.ek = ((R * (self.T + 273.15)) / F) * math.log(
                    ko / ki
                )  # Manual Calculation of ek in order to use Schild F and R values
                nao = 154.0  # [mM] External Na Concentration
                nai = 8.9  # [mM] Internal Na Concentration
                nastyle = neuron.h.ion_style(
                    "na_ion", 1, 2, 0, 0, 0
                )  # Allows ena to be calculated manually
                sec.ena = ((R * (self.T + 273.15)) / F) * math.log(
                    nao / nai
                )  # Manual Calculation of ena in order to use Schild F and R values
                if self.model == "Schild_97":
                    sec.gbar_naf97mean = 0.022434928  # [S/cm^2] This block sets the conductance to the conductances in Schild 1997
                    sec.gbar_nas97mean = 0.022434928
                    sec.gbar_kd = 0.001956534
                    sec.gbar_ka = 0.001304356
                    sec.gbar_kds = 0.000782614
                    sec.gbar_kca = 0.000913049
                    sec.gbar_can = 0.000521743
                    sec.gbar_cat = 0.00018261
                    sec.gbna_leakSchild = 1.8261e-05
                    sec.gbca_leakSchild = 9.13049e-06
                sec.Ra = 100
                sec.cm = 1.326291192
                sec.v = -48
                sec.L_caintscale = self.L / self.Nsec
                sec.L_caextscale = self.L / self.Nsec

    ###############################
    ## Intracellular stimulation ##
    ###############################
    def insert_I_Clamp(self, position, t_start, duration, amplitude):
        """
        Insert a I clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the axon
        t_start     : float
            starting time, in ms
        duration    : float
            duration of the pulse, in ms
        amplitude   : float
            amplitude of the pulse (nA)
        """
        # adapt position to the number of sections
        portion_length = 1.0 / self.Nsec
        stim_sec = int(math.floor(position / portion_length))
        stim_pos = (position / portion_length) - math.floor(position / portion_length)
        # add the stimulation to the axon
        self.intra_current_stim.append(
            neuron.h.IClamp(stim_pos, sec=self.unmyelinated_sections[stim_sec])
        )
        # modify the stimulation parameters
        self.intra_current_stim[-1].delay = t_start
        self.intra_current_stim[-1].dur = duration
        self.intra_current_stim[-1].amp = amplitude
        # save the stimulation parameter for results
        self.intra_current_stim_positions.append(position * self.L)
        self.intra_current_stim_starts.append(t_start)
        self.intra_current_stim_durations.append(duration)
        self.intra_current_stim_amplitudes.append(amplitude)

    def insert_V_Clamp(self, position, stimulus):
        """
        Insert a V clamp stimulation

        Parameters
        ----------
        position    : float
            relative position over the axon
        stimulus    : stimulus object
            stimulus for the clamp, see Stimulus.py for more information
        """
        # adapt position to the number of sections
        portion_length = 1.0 / self.Nsec
        stim_sec = int(math.floor(position / portion_length))
        stim_pos = (position / portion_length) - math.floor(position / portion_length)
        # add the stimulation to the axon
        self.intra_voltage_stim = neuron.h.VClamp(
            stim_pos, sec=self.unmyelinated_sections[stim_sec]
        )
        # save the stimulation parameter for results
        self.intra_current_stim_position = position * self.L
        # save the stimulus for later use
        self.intra_voltage_stim_stimulus = stimulus
        # set fake duration
        self.intra_voltage_stim.dur[0] = 1e9

    ##############################
    ## Result recording methods ##
    ##############################
    def __set_recorders_with_key(self, *args):
        """
        To automate the methods set_recorder. For internal use only.
        Parameters
        ----------
        *args    : list(tuple)
            list of tuple containing a rec list to set and the corresonding key to access
            NB: keys should be str such as "_ref_xxx_yyy" where xxx is the variable to access
            and yyy the .mod file suffix if the variable is in one
        """
        for k in range(self.Nsec):
            for pos in self.rec_position_list[k]:
                for t in args:
                    key = t[1]
                    # print(dir(self.unmyelinated_sections[k](pos)))
                    # print(key, getattr(self.unmyelinated_sections[k](pos),key))
                    rec = neuron.h.Vector().record(
                        getattr(self.unmyelinated_sections[k](pos), key),
                        sec=self.unmyelinated_sections[k],
                    )
                    t[0].append(rec)

    def __get_var_from_mod(
        self, key
    ):
        """
        return a column with value in every recording point of a constant from a mod. For internal use only.
        """
        val = np.zeros((len(self.x_rec)))
        i = 0
        for k in range(self.Nsec):
            for pos in self.rec_position_list[k]:
                #print(getattr(self.unmyelinated_sections[k](pos), key)[0])
                val[i] = getattr(self.unmyelinated_sections[k](pos), key)[0]
                i += 1
        return val

    def __get_recorders_from_list(self, reclist):
        """
        Convert reclist in np.array To automate methods set_recorder. For internal use only.
        Parameters
        ----------
        reclist     : neuron.h.List
            List in witch the reccorders are saved

        Returns
        -------
        val         : np.array
            array of every recorded value for all rec point and time
        """
        dim = (self.Nrec, self.t_len)
        val = np.zeros(dim)
        for k in range(dim[0]):
            val[k, :] = np.asarray(reclist[k])
        return val

    def set_membrane_voltage_recorders(self):
        """
        Prepare the membrane voltage recording. For internal use only.
        """
        self.vreclist = neuron.h.List()
        self.__set_recorders_with_key((self.vreclist, "_ref_v"))

    def get_membrane_voltage(self):
        """
        get the membrane voltage at the end of simulation. For internal use only.
        """
        return self.__get_recorders_from_list(self.vreclist)

    def set_membrane_current_recorders(self):
        """
        Prepare the membrane current recording. For internal use only.
        """
        self.ireclist = neuron.h.List()
        self.__set_recorders_with_key((self.ireclist, "_ref_i_membrane"))

    def get_membrane_current(self):
        """
        get the membrane current at the end of simulation. For internal use only.
        """
        return self.__get_recorders_from_list(self.ireclist)

    def set_ionic_current_recorders(self):
        """
        Prepare the ionic currents recording. For internal use only.
        """
        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            self.i_na_reclist = neuron.h.List()
            self.i_k_reclist = neuron.h.List()
            self.i_l_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                (self.i_na_reclist, "_ref_nai"),
                (self.i_k_reclist, "_ref_ki"),
                (self.i_l_reclist, "_ref_i_pas"),
            )
        else:
            self.i_na_reclist = neuron.h.List()
            self.i_k_reclist = neuron.h.List()
            self.i_ca_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                (self.i_na_reclist, "_ref_nai"),
                (self.i_k_reclist, "_ref_ki"),
                (self.i_ca_reclist, "_ref_cai"),
            )

    def get_ionic_current(self):
        """
        get the ionic currents at the end of simulation. For internal use only.
        """
        results = []
        results += [self.__get_recorders_from_list(self.i_na_reclist)]
        results += [self.__get_recorders_from_list(self.i_k_reclist)]
        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            results += [self.__get_recorders_from_list(self.i_l_reclist)]
        else:
            results += [self.__get_recorders_from_list(self.i_ca_reclist)]
        return results

    def set_conductance_recorders(self):
        """
        Prepare the membrane conductance recording. For internal use only.
        """
        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            self.g_na_reclist = neuron.h.List()
            self.g_k_reclist = neuron.h.List()
            self.g_l_reclist = neuron.h.List()
            if self.model == "HH":
                self.__set_recorders_with_key(
                    (self.g_na_reclist, "_ref_gna_hh"),
                    (self.g_k_reclist, "_ref_gk_hh"),
                    (self.g_l_reclist, "_ref_gl_hh"),
                )
            elif self.model == "Rattay_Aberham":
                self.__set_recorders_with_key(
                    (self.g_na_reclist, "_ref_gna_RattayAberham"),
                    (self.g_k_reclist, "_ref_gk_RattayAberham"),
                    (self.g_l_reclist, "_ref_gl_RattayAberham"),
                )
            else:
                self.__set_recorders_with_key(
                    (self.g_na_reclist, "_ref_gna_nahh"),
                    (self.g_k_reclist, "_ref_gkdr_borgkdr"),
                    (self.g_l_reclist, "_ref_g_pas"),
                )
        elif self.model == "Tigerholm":
            self.g_nav17_reclist = neuron.h.List()
            self.g_nav18_reclist = neuron.h.List()
            self.g_nav19_reclist = neuron.h.List()
            self.g_kA_reclist = neuron.h.List()
            self.g_kM_reclist = neuron.h.List()
            self.g_kdr_reclist = neuron.h.List()
            self.g_kna_reclist = neuron.h.List()
            self.g_h_reclist = neuron.h.List()
            self.g_naleak_reclist = neuron.h.List()
            self.g_kleak_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                (self.g_nav17_reclist, "_ref_g_nattxs"),
                (self.g_nav18_reclist, "_ref_g_nav1p8"),
                (self.g_nav19_reclist, "_ref_g_nav1p9"),
                (self.g_kA_reclist, "_ref_g_ks"),
                (self.g_kM_reclist, "_ref_g_kf"),
                (self.g_kdr_reclist, "_ref_g_kdrTiger"),
                (self.g_kna_reclist, "_ref_g_kna"),
                (self.g_h_reclist, "_ref_g_h"),
                (self.g_naleak_reclist, "_ref_gnaleak_leak"),
                (self.g_kleak_reclist, "_ref_gkleak_leak"),
            )
        else:
            self.g_naf_reclist = neuron.h.List()
            self.g_nas_reclist = neuron.h.List()
            self.g_kd_reclist = neuron.h.List()
            self.g_ka_reclist = neuron.h.List()
            self.g_kds_reclist = neuron.h.List()
            self.g_kca_reclist = neuron.h.List()
            self.g_can_reclist = neuron.h.List()
            self.g_cat_reclist = neuron.h.List()
            sup_key = ""
            if self.model == "Schild_97":
                sup_key = "97mean"
            self.__set_recorders_with_key(
                (self.g_naf_reclist, "_ref_g_naf" + sup_key),
                (self.g_nas_reclist, "_ref_g_nas" + sup_key),
                (self.g_kd_reclist, "_ref_g_kd"),
                (self.g_ka_reclist, "_ref_g_ka"),
                (self.g_kds_reclist, "_ref_g_kds"),
                (self.g_kca_reclist, "_ref_g_kca"),
                (self.g_can_reclist, "_ref_g_can"),
                (self.g_cat_reclist, "_ref_g_cat"),
            )

    def get_membrane_conductance(self):
        """
        get the membrane voltage at the end of simulation. For internal use only.
        NB: [S/cm^{2}] (see Neuron unit)
        """
        return sum(self.get_ionic_conductance())

    def get_ionic_conductance(self):
        """
        get the membrane conductance at the end of simulation. For internal use only.
        NB: [S/cm^{2}] (see Neuron unit)
        """
        results = []
        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            results += [self.__get_recorders_from_list(self.g_na_reclist)]
            results += [self.__get_recorders_from_list(self.g_k_reclist)]
            results += [self.__get_recorders_from_list(self.g_l_reclist)]
        elif self.model == "Tigerholm":
            results += [self.__get_recorders_from_list(self.g_nav17_reclist)]
            results += [self.__get_recorders_from_list(self.g_nav18_reclist)]
            results += [self.__get_recorders_from_list(self.g_nav19_reclist)]
            results += [self.__get_recorders_from_list(self.g_kA_reclist)]
            results += [self.__get_recorders_from_list(self.g_kM_reclist)]
            results += [self.__get_recorders_from_list(self.g_kdr_reclist)]
            results += [self.__get_recorders_from_list(self.g_kna_reclist)]
            results += [self.__get_recorders_from_list(self.g_h_reclist)]
            results += [self.__get_recorders_from_list(self.g_naleak_reclist)]
            results += [self.__get_recorders_from_list(self.g_kleak_reclist)]
        else:
            results += [self.__get_recorders_from_list(self.g_naf_reclist)]
            results += [self.__get_recorders_from_list(self.g_nas_reclist)]
            results += [self.__get_recorders_from_list(self.g_kd_reclist)]
            results += [self.__get_recorders_from_list(self.g_ka_reclist)]
            results += [self.__get_recorders_from_list(self.g_kds_reclist)]
            results += [self.__get_recorders_from_list(self.g_kca_reclist)]
            results += [self.__get_recorders_from_list(self.g_can_reclist)]
            results += [self.__get_recorders_from_list(self.g_cat_reclist)]
        return results

    def get_membrane_capacitance(self):
        """
        get the membrane capacitance
        NB: [uF/cm^{2}] (see Neuron unit)
        """
        return self.__get_var_from_mod('_ref_cm')

    def set_particules_values_recorders(self):
        """
        Prepare the particule value recording. For internal use only.
        """

        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            self.hhmreclist = neuron.h.List()
            self.hhnreclist = neuron.h.List()
            self.hhhreclist = neuron.h.List()
            if self.model == "HH":
                self.__set_recorders_with_key(
                    (self.hhmreclist, "_ref_m_hh"),
                    (self.hhnreclist, "_ref_n_hh"),
                    (self.hhhreclist, "_ref_h_hh"),
                )
            elif self.model == "Rattay_Aberham":
                self.__set_recorders_with_key(
                    (self.hhmreclist, "_ref_m_RattayAberham"),
                    (self.hhnreclist, "_ref_n_RattayAberham"),
                    (self.hhhreclist, "_ref_h_RattayAberham"),
                )
            else:
                self.__set_recorders_with_key(
                    (self.hhmreclist, "_ref_m_nahh"),
                    (self.hhnreclist, "_ref_n_borgkdr"),
                    (self.hhhreclist, "_ref_h_nahh"),
                )
        elif self.model == "Tigerholm":
            # NAV 1.8
            self.m_nav18_reclist = neuron.h.List()
            self.h_nav18_reclist = neuron.h.List()
            self.s_nav18_reclist = neuron.h.List()
            self.u_nav18_reclist = neuron.h.List()
            # NAV 1.9
            self.m_nav19_reclist = neuron.h.List()
            self.h_nav19_reclist = neuron.h.List()
            self.s_nav19_reclist = neuron.h.List()
            # NATTX - sensitive
            self.m_nattxs_reclist = neuron.h.List()
            self.h_nattxs_reclist = neuron.h.List()
            self.s_nattxs_reclist = neuron.h.List()
            # K delayed rectifier
            self.n_kdr_reclist = neuron.h.List()
            # K fast channel
            self.m_kf_reclist = neuron.h.List()
            self.h_kf_reclist = neuron.h.List()
            # K slow channel
            self.ns_ks_reclist = neuron.h.List()
            self.nf_ks_reclist = neuron.h.List()
            # Sodium dependent K channel
            self.w_kna_reclist = neuron.h.List()
            # Hyperpolarization channel
            self.ns_h_reclist = neuron.h.List()
            self.nf_h_reclist = neuron.h.List()

            self.__set_recorders_with_key(
                (self.m_nav18_reclist, "_ref_m_nav1p8"),
                (self.h_nav18_reclist, "_ref_h_nav1p8"),
                (self.s_nav18_reclist, "_ref_s_nav1p8"),
                (self.u_nav18_reclist, "_ref_u_nav1p8"),
                (self.m_nav19_reclist, "_ref_m_nav1p9"),
                (self.h_nav19_reclist, "_ref_h_nav1p9"),
                (self.s_nav19_reclist, "_ref_s_nav1p9"),
                (self.m_nattxs_reclist, "_ref_m_nattxs"),
                (self.h_nattxs_reclist, "_ref_h_nattxs"),
                (self.s_nattxs_reclist, "_ref_s_nattxs"),
                (self.n_kdr_reclist, "_ref_n_kdrTiger"),
                (self.m_kf_reclist, "_ref_m_kf"),
                (self.h_kf_reclist, "_ref_h_kf"),
                (self.ns_ks_reclist, "_ref_ns_ks"),
                (self.nf_ks_reclist, "_ref_nf_ks"),
                (self.w_kna_reclist, "_ref_w_kna"),
                (self.ns_h_reclist, "_ref_ns_h"),
                (self.nf_h_reclist, "_ref_nf_h"),
            )
        else:  # should be both Schild_94 or Schild_97
            # High Threshold long lasting Ca
            self.d_can_reclist = neuron.h.List()
            self.f1_can_reclist = neuron.h.List()
            self.f2_can_reclist = neuron.h.List()
            # Low Threshold transient Ca
            self.d_cat_reclist = neuron.h.List()
            self.f_cat_reclist = neuron.h.List()
            # Early Transient Outward K
            self.p_ka_reclist = neuron.h.List()
            self.q_ka_reclist = neuron.h.List()
            # Ca activated K
            self.c_kca_reclist = neuron.h.List()
            # Delayed rectifier K
            self.n_kd_reclist = neuron.h.List()
            # Slowly inactivated K
            self.x_kds_reclist = neuron.h.List()
            self.y1_kds_reclist = neuron.h.List()
            self.__set_recorders_with_key(
                (self.d_can_reclist, "_ref_d_can"),
                (self.f1_can_reclist, "_ref_f1_can"),
                (self.f2_can_reclist, "_ref_f2_can"),
                (self.d_cat_reclist, "_ref_d_cat"),
                (self.f_cat_reclist, "_ref_f_cat"),
                (self.p_ka_reclist, "_ref_p_ka"),
                (self.q_ka_reclist, "_ref_q_ka"),
                (self.c_kca_reclist, "_ref_c_kca"),
                (self.n_kd_reclist, "_ref_n_kd"),
                (self.x_kds_reclist, "_ref_x_kds"),
                (self.y1_kds_reclist, "_ref_y1_kds"),
            )
            if self.model == "Schild_94":
                # Fast Na
                self.m_naf_reclist = neuron.h.List()
                self.h_naf_reclist = neuron.h.List()
                self.l_naf_reclist = neuron.h.List()
                # Slow Na
                self.m_nas_reclist = neuron.h.List()
                self.h_nas_reclist = neuron.h.List()
                self.__set_recorders_with_key(
                    (self.m_naf_reclist, "_ref_m_naf"),
                    (self.h_naf_reclist, "_ref_h_naf"),
                    (self.l_naf_reclist, "_ref_l_naf"),
                    (self.m_nas_reclist, "_ref_m_nas"),
                    (self.h_nas_reclist, "_ref_h_nas"),
                )
            else:  # should be Schild_94
                # Fast Na
                self.m_naf_reclist = neuron.h.List()
                self.h_naf_reclist = neuron.h.List()
                # Slow Na
                self.m_nas_reclist = neuron.h.List()
                self.h_nas_reclist = neuron.h.List()
                self.__set_recorders_with_key(
                    (self.m_naf_reclist, "_ref_m_naf97mean"),
                    (self.h_naf_reclist, "_ref_h_naf97mean"),
                    (self.m_nas_reclist, "_ref_m_nas97mean"),
                    (self.h_nas_reclist, "_ref_h_nas97mean"),
                )

    def get_particles_values(self):
        """
        get the particules values at the end of simulation. For internal use only.
        """
        results = []
        if self.model in ["HH", "Rattay_Aberham", "Sundt"]:
            results += [self.__get_recorders_from_list(self.hhmreclist)]
            results += [self.__get_recorders_from_list(self.hhnreclist)]
            results += [self.__get_recorders_from_list(self.hhhreclist)]
        elif self.model in ["Tigerholm"]:
            results += [self.__get_recorders_from_list(self.m_nav18_reclist)]
            results += [self.__get_recorders_from_list(self.h_nav18_reclist)]
            results += [self.__get_recorders_from_list(self.s_nav18_reclist)]
            results += [self.__get_recorders_from_list(self.u_nav18_reclist)]
            results += [self.__get_recorders_from_list(self.m_nav19_reclist)]
            results += [self.__get_recorders_from_list(self.h_nav19_reclist)]
            results += [self.__get_recorders_from_list(self.s_nav19_reclist)]
            results += [self.__get_recorders_from_list(self.m_nattxs_reclist)]
            results += [self.__get_recorders_from_list(self.h_nattxs_reclist)]
            results += [self.__get_recorders_from_list(self.s_nattxs_reclist)]
            results += [self.__get_recorders_from_list(self.n_kdr_reclist)]
            results += [self.__get_recorders_from_list(self.m_kf_reclist)]
            results += [self.__get_recorders_from_list(self.h_kf_reclist)]
            results += [self.__get_recorders_from_list(self.ns_ks_reclist)]
            results += [self.__get_recorders_from_list(self.nf_ks_reclist)]
            results += [self.__get_recorders_from_list(self.w_kna_reclist)]
            results += [self.__get_recorders_from_list(self.ns_h_reclist)]
            results += [self.__get_recorders_from_list(self.nf_h_reclist)]

        else:  # should be "Schild_94" or "Schild_97"
            results += [self.__get_recorders_from_list(self.d_can_reclist)]
            results += [self.__get_recorders_from_list(self.f1_can_reclist)]
            results += [self.__get_recorders_from_list(self.f2_can_reclist)]
            results += [self.__get_recorders_from_list(self.d_cat_reclist)]
            results += [self.__get_recorders_from_list(self.f_cat_reclist)]
            results += [self.__get_recorders_from_list(self.p_ka_reclist)]
            results += [self.__get_recorders_from_list(self.q_ka_reclist)]
            results += [self.__get_recorders_from_list(self.c_kca_reclist)]
            results += [self.__get_recorders_from_list(self.n_kd_reclist)]
            results += [self.__get_recorders_from_list(self.x_kds_reclist)]
            results += [self.__get_recorders_from_list(self.y1_kds_reclist)]
            results += [self.__get_recorders_from_list(self.m_naf_reclist)]
            results += [self.__get_recorders_from_list(self.h_naf_reclist)]
            results += [self.__get_recorders_from_list(self.m_nas_reclist)]
            results += [self.__get_recorders_from_list(self.h_nas_reclist)]
            if self.model == "Schild_94":
                results += [self.__get_recorders_from_list(self.l_naf_reclist)]
        return results



# pylint: skip-file
""" NeuRon Virtualizer, large scale modeling of Peripheral Nervous System with random stimulation waveforms"""

__copyright__ = "2023, Florian Kolbl"
__license__ = "CeCILL"
__version__ = "0.9.15"
__title__ = "NeuRon Virtualizer"
__authors__ = "Florian Kolbl, Roland Giraud, Louis Regnacq, Thomas Couppey"
__contributors__ = __authors__
__project__ = "NeuRon Virtualizer (NRV)"

#####################################
#  check environnement variables    #
#  check correct NRV configuration  #
#####################################
import os
import inspect
import platform
# GMSH must be imported before neuron to prevent installation issues
import gmsh


# create a dummy object to locate frameworks path
class DummyClass:
    """Dummy class"""

    pass


nrv_path = os.path.dirname(os.path.abspath(inspect.getsourcefile(DummyClass)))
root_path = nrv_path.replace("/nrv/", "")
# check the PATH with os.environ['PATH'], modify bash/zsh profile
current_PATH = os.environ["PATH"]
conf_file = ".bashrc"
if platform.system() == "Darwin":
    # for MacOsX platforms
    conf_file = ".zshrc"
if not (
    nrv_path + ":" in current_PATH
    or ":" + nrv_path in current_PATH
    or "/nrv:" in current_PATH
):
    with open(os.path.expanduser("~/" + conf_file), "r") as outfile:
        # test if bash/zsh profile already modified
        Flag = False
        lines = outfile.readlines()
        for line in lines:
            if "NRV setup" in line:
                Flag = True
    outfile.close()
    with open(os.path.expanduser("~/" + conf_file), "a") as outfile:
        # modify or suggest user to source it
        if not Flag:
            # in this case, thte bash/zsh profile is not modified yet,
            # # then do it
            outfile.write("\n\n\n# >>>>> NRV setup >>>>>\n")
            outfile.write('export PATH="' + nrv_path + ':$PATH"\n')
            outfile.write("# <<<<< NRV setup <<<<<\n")
            print(
                conf_file
                + " file modified, please source or restart console to be able to used nrv2calm"
            )
        # ask to source bash/zsh profile
        else:
            print(
                "Please restart console to be able to use nrv2calm or source your own bash/zsh profile"
            )
    outfile.close()
# create the environnement variable NRVPATH if it does not exist
if "NRVPATH" not in os.environ:
    os.environ["NRVPATH"] = nrv_path
# change the permissions on nrv2calm
if not os.access(nrv_path + "/nrv2calm", os.X_OK):
    mode = os.stat(nrv_path + "/nrv2calm").st_mode
    mode |= (mode & 0o444) >> 2  # copy R bits to X
    os.chmod(nrv_path + "/nrv2calm", mode)

######################
#  Public interface  #
######################
from .backend import compileMods
from .backend.parameters import *
from .backend.NRV_Class import load_any

from .fmod.materials import *
from .fmod.electrodes import *
from .fmod.stimulus import *
from .fmod.extracellular import *
from .fmod.recording import *
from .fmod.FEM.FEM import *
from .fmod.FEM.COMSOL_model import *
from .fmod.FEM.FENICS_model import *

######### May not be requiered at the end ###############
from .fmod.FEM.mesh_creator.MshCreator import *
from .fmod.FEM.mesh_creator.NerveMshCreator import *
from .fmod.FEM.mesh_creator.NRV_Msh import *
from .fmod.FEM.fenics_utils.FEMSimulation import *
from .fmod.FEM.fenics_utils.SimParameters import *
from .fmod.FEM.fenics_utils.SimResult import *

########################################################

from .nmod.axons import *
from .nmod.unmyelinated import *
from .nmod.myelinated import *
from .nmod.fascicles import *
from .nmod.fascicle_generator import *
from .nmod.nerve import *

from .utils.saving_handler import *
from .utils.nrv_function import *
from .utils.cell.CL_postprocessing import *
from .utils.cell.CL_simulations import *
from .utils.cell.CL_discretization import *
from .utils.fascicle.FL_postprocessing import *

from .optim.CostFunctions import *
from .optim.Optimizers import *
from .optim.Problems import *
from .optim.optim_utils.ContextModifier import *
from .optim.optim_utils.CostEvaluation import *
from .optim.optim_utils.InterpolationFunctions import *


nrv.rasterize(axon_sim, "V_mem")
nrv.filter_freq(axon_sim, "V_mem", 10)

plt.figure()
for node in range(len(axon_sim["x_rec"])):
    plt.plot(
        axon_sim["t"], axon_sim["V_mem_filtered"][node] + 100 * node, color="k"
    )
plt.xlabel("time (ms)")
plt.yticks([])
plt.ylabel("membrane voltage at Nodes of Ranvier allong axon")
title = (
    "Axon "
    + str(k)
    + ", myelination is "
    + str(axon_sim["myelinated"])
    + ", "
    + str(axon_sim["diameter"])
    + " um diameter"
)
plt.title(title)
plt.xlim(0, axon_sim["tstop"])
plt.tight_layout()
if self.save_results:
    fig_name = folder_name + "/Vmem_axon_" + str(k) + ".pdf"
    plt.savefig(fig_name)
    plt.close()

nrv.remove_key(axon_sim, "V_mem", verbose=self.verbose)
nrv.remove_key(axon_sim, "V_mem_filtered", verbose=self.verbose)



import gmsh
import numpy as np
import scipy
from dolfinx.fem import Expression, Function, FunctionSpace
from dolfinx.geometry import (
    BoundingBoxTree,
    compute_colliding_cells,
    compute_collisions,
)
from dolfinx.io.gmshio import model_to_mesh
from dolfinx.io.utils import XDMFFile
from mpi4py import MPI

from ....backend.file_handler import rmv_ext
from ....backend.log_interface import rise_error, rise_warning
from ....backend.MCore import MCH, synchronize_processes
from ....backend.NRV_Class import NRV_class
from ..mesh_creator.MshCreator import (
    is_MshCreator,
    clear_gmsh,
)


###############
## Functions ##
###############
def is_sim_res(result):
    """
    check if an object is a SimResult, return True if yes, else False

    Parameters
    ----------
    result : object
        object to test

    Returns
    -------
    bool
        True it the type is a SimResult object
    """
    return isinstance(result, SimResult)


def save_sim_res_list(sim_res_list, fname, dt=1):
    """
    save a list of SimResults in a xdmf file
    """
    fname = rmv_ext(fname) + ".xdmf"
    N_list = len(sim_res_list)
    xdmf = XDMFFile(sim_res_list[0].domain.comm, fname, "w")
    xdmf.write_mesh(sim_res_list[0].domain)
    for E in range(N_list):
        xdmf.write_function(sim_res_list[E].vout, E * dt)


def read_gmsh(mesh, comm=MPI.COMM_WORLD, rank=0, gdim=3):
    """
    Given a mesh_file or a MeshCreator returns mesh cell_tags and facet_tags
    (see dolfinx.io.gmshio.model_to_mesh for more details)
    NB: copy of dolfinx.io.gmshio.read_from_msh verbose from gmsh

    Parameters
    ----------
    mesh_file : str,
        Path and name of mesh file
        NB: extention is not required but must be a .msh file
    comm            : tupple (str, int)
        The MPI communicator to use for mesh creation
    rank            : tupple (str, int)
        The rank the Gmsh model is initialized on
    gdim            : int
        dimension of the mesh, by default 3

    Returns
    -------
    output  :  tuple(3)
        (domain, cell_tag, facet_tag)
    """
    if isinstance(mesh, str):
        mesh_file = rmv_ext(mesh) + ".msh"
        clear_gmsh()
        gmsh.initialize()
        gmsh.option.setNumber("General.Verbosity", 2)
        gmsh.model.add("Mesh from file")
        gmsh.merge(mesh_file)
        output = model_to_mesh(gmsh.model, comm=comm, rank=rank, gdim=gdim)
        gmsh.finalize()
    elif is_MshCreator(mesh):
        output = model_to_mesh(mesh.model, comm=comm, rank=rank, gdim=gdim)
    else:
        rise_error("mesh should be either a filename or a MeshCreator")
    return output


def domain_from_meshfile(mesh_file):
    """
        return only the domain from mesh_file

    Parameters
    ----------
    mesh_file : str
        Path and name of mesh file
        NB: extention is not required but must be a .msh file

    Returns
    -------
        output  :
            domain of the mesh contain in the mesh_file
    """
    return read_gmsh(mesh_file)[0]


def V_from_meshfile(mesh_file, elem=("Lagrange", 1)):
    mesh = domain_from_meshfile(mesh_file)
    V = FunctionSpace(mesh, elem)
    return V


class SimResult(NRV_class):
    """
    Result of a FEMSimulation.
    Store the resulting function space giving the possibility to apply basic mathematical
    operations on multiple results
    """

    def __init__(
        self,
        mesh_file="",
        domain=None,
        elem=("Lagrange", 1),
        V=None,
        vout=None,
        comm=MPI.COMM_WORLD,
    ):
        """
        initialisation of the SimParameters:

        Parameters
        ----------
        mesh_file       :str
            mesh directory and file name: by default ""
        domain       : None or mesh
            mesh domain on which the result is defined, by default None
        elem            :tupple (str, int)
            if None, ("Lagrange", 1), else (element type, element order), by default None
        V       : None or dolfinx.fem.FunctionSpace
            FunctionSpace on which the result is defined, by default None
        vout       : None or dolfinx.fem.Function
            Function resulting from the FEMSimulation, by default None
        comm            :int
            The MPI communicator to use for mesh creation, by default MPI.COMM_WORLD
        """
        super().__init__()
        self.type = "simresult"
        self.mesh_file = mesh_file

        self.domain = domain
        self.V = V
        self.elem = elem
        self.vout = vout
        self.comm = comm

    def set_sim_result(
        self, mesh_file="", domain=None, V=None, elem=None, vout=None, comm=None
    ):
        if mesh_file != "":
            self.mesh_file = mesh_file
        if domain is not None:
            self.domain = domain
        if elem is not None:
            self.elem = elem
        if V is not None:
            self.V = V
        if vout is not None:
            self.vout = vout
        self.comm = comm

    def save(self, file, ftype="xdmf", overwrite=True):
        if ftype == "xdmf":
            fname = rmv_ext(file) + ".xdmf"
            with XDMFFile(self.comm, fname, "w") as file:
                if not overwrite:
                    file.parameters.update(
                        {"functions_share_mesh": True, "rewrite_function_mesh": False}
                    )
                else:
                    file.write_mesh(self.domain)
                file.write_function(self.vout)
        else:
            fname = rmv_ext(file) + ".sres"
            mdict = {
                "mesh_file": self.mesh_file,
                "element": self.elem,
                "vout": self.vout.vector[:],
            }
            scipy.io.savemat(fname, mdict)
            return mdict

    def load(self, file):
        fname = rmv_ext(file) + ".sres"
        mdict = scipy.io.loadmat(fname)
        self.mesh_file = mdict["mesh_file"][0]
        self.elem = (mdict["element"][0].strip(), int(mdict["element"][1]))
        if self.domain is None:
            self.domain = domain_from_meshfile(self.mesh_file)
            self.V = FunctionSpace(self.domain, self.elem)
        self.vout = Function(self.V)
        self.vout.vector[:] = mdict["vout"]

    def save_sim_result(self, file, ftype="xdmf", overwrite=True):
        rise_warning("save_sim_result is a deprecated method use save")
        self.save(file=file, ftype=ftype, overwrite=overwrite)

    def load_sim_result(self, data="sim_result.json"):
        rise_warning("load_sim_result is a deprecated method use load")
        # self.load(data=data)
        # FIXME: data does not exist yet in load method
        self.load(file=data)

    #############
    ## methods ##
    #############

    def vector(self):
        return self.vout.x.array

    def aline_V(self, res2):
        """
        Change the function space of the result to aline with the result of another SimResult

        Parameters
        ----------
        res2 : SimResult
            result to aline with
        """
        if is_sim_res(res2):
            if self.mesh_file == res2.mesh_file:
                expr = Expression(self.vout, res2.V.element.interpolation_points())
                self.vout = Function(res2.V)
                self.vout.interpolate(expr)
                self.V = res2.V
            else:
                rise_error(
                    "To aline mesh function reslults must have the same meshfile"
                )
        else:
            rise_error("Mesh function alinment must be done with SimResult")

    def eval(self, X, is_multi_proc=False):
        """
        Eval the result field at X position
        """
        N = len(X)
        to_round = False
        cells = []
        points_on_proc = []
        tree = BoundingBoxTree(self.domain, self.domain.geometry.dim)
        # Find cells whose bounding-box collide with the the points
        cells_candidates = compute_collisions(tree, X)
        # Choose one of the cells that contains the point
        cells_colliding = compute_colliding_cells(self.domain, cells_candidates, X)
        for i in range(N):
            cell = cells_colliding.links(i)
            if is_multi_proc:
                if len(cell) > 0:
                    points_on_proc.append(X[i])
                    cells.append(cells_colliding.links(i)[0])
            else:
                if len(cell) == 0:
                    if i == N - 1 and i > 0:
                        to_round = True
                    else:
                        rise_warning(
                            X[i], " not found in mesh, value of ", X[i - 1], " reused"
                        )
                    cells += [cells[-1]]
                else:
                    cells += [cell[0]]

        if is_multi_proc:
            points_on_proc = np.array(points_on_proc, dtype=np.float64)
            s_values = self.vout.eval(points_on_proc, cells)
            if len(points_on_proc) > 0:
                m_values = np.concatenate((points_on_proc.T, s_values.T)).T
            else:
                m_values = np.array([])
            m_values = self.comm.gather(m_values.tolist(), root=0)
            if MCH.do_master_only_work():
                val = []
                for i in range(len(m_values)):
                    for j in range(len(m_values[i])):
                        val += [m_values[i][j]]
                val = np.array(val)
                values = np.empty((N), dtype="float64")
                for i, p in enumerate(X):
                    i_p = np.where((np.isclose(val[:, :3], p)).all(axis=1))[0]
                    if len(i_p > 0):
                        values[i] = val[i_p[0], 3]
                    else:
                        values[i] = values[i - 1]
            else:
                values = np.empty((N), dtype="float64")
            synchronize_processes()
            self.comm.Bcast(values, root=0)
        else:
            values = self.vout.eval(X, cells)
            if N > 1:
                values = values[:, 0]
        if to_round:
            X_1 = X[-1]
            X_1[0] = round(X_1[0], 1)
            value_1 = self.eval([X_1], is_multi_proc=is_multi_proc)
            values[-1] = value_1[0]
        return values

    #####################
    ## special methods ##
    #####################
    def __abs__(self):
        res = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        if self.vout is not None:
            expr = Expression(abs(self.vout), self.V.element.interpolation_points())
            res.vout = Function(self.V)
            self.vout.interpolate(expr)
        return res

    def __neg__(self):
        res = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        if self.vout is not None:
            expr = Expression(-1 * self.vout, self.V.element.interpolation_points())
            res.vout = Function(self.V)
            self.vout.interpolate(expr)
        return res

    def __add__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout + b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout + b, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __sub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout - b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout - b, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __mul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(self.vout * b.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(self.vout * b, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __radd__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout + self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b + self.vout, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __rsub__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout - self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b - self.vout, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __rmul__(self, b):
        if is_sim_res(b):
            self.aline_V(b)
            expr = Expression(b.vout * self.vout, self.V.element.interpolation_points())
        else:
            expr = Expression(b * self.vout, self.V.element.interpolation_points())

        C = SimResult(
            mesh_file=self.mesh_file,
            V=self.V,
            domain=self.domain,
            elem=self.elem,
            comm=self.comm,
        )
        C.vout = Function(self.V)
        C.vout.interpolate(expr)
        return C

    def __eq__(self, b):
        if is_sim_res(b):
            if b.mesh_file == self.mesh_file:
                if np.allclose(self.vector(), b.vector()):
                    return True
        return False

    def __ne__(self, b):  # self != b
        return not self == b
